# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import abc
import itertools
from typing import Callable, Dict, Generic, List, Mapping, Optional, Type, Union

import attrs
import h5py

from acconeer.exptool import a121
from acconeer.exptool.a121._core import utils as core_utils
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._base import (
    GenericProcessorBase,
    InputT,
    MetadataT,
    ProcessorConfigT,
    ResultT,
)
from acconeer.exptool.a121.algo._plugins._a121 import A121BackendPluginBase
from acconeer.exptool.app.new import (
    PluginGeneration,
)
from acconeer.exptool.app.new.backend import (
    BackendLogger,
    GeneralMessage,
    Message,
    PlotMessage,
    RecipientLiteral,
    StatusMessage,
    is_task,
)


@attrs.mutable(kw_only=True)
class ProcessorPluginPreset(Generic[ProcessorConfigT]):
    session_config: a121.SessionConfig = attrs.field()
    processor_config: ProcessorConfigT = attrs.field()


@attrs.mutable(kw_only=True)
class ProcessorBackendPluginSharedState(Generic[ProcessorConfigT]):
    session_config: a121.SessionConfig = attrs.field()
    processor_config: ProcessorConfigT = attrs.field()
    replaying: bool = attrs.field(default=False)
    metadata: Union[a121.Metadata, list[dict[int, a121.Metadata]], None] = attrs.field(
        default=None
    )

    @property
    def ready(self) -> bool:
        try:
            self.session_config.validate()
            self.processor_config.validate(self.session_config)
        except a121.ValidationError:
            return False
        else:
            return True


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage, Generic[ProcessorConfigT]):
    session_config: a121.SessionConfig
    processor_config: ProcessorConfigT
    metadata: Union[a121.Metadata, list[dict[int, a121.Metadata]]]
    name: str = attrs.field(default="setup", init=False)
    recipient: RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class GenericProcessorBackendPluginBase(
    Generic[InputT, ProcessorConfigT, ResultT, MetadataT],
    A121BackendPluginBase[ProcessorBackendPluginSharedState[ProcessorConfigT]],
):
    _processor_instance: Optional[GenericProcessorBase[InputT, ResultT]]
    _started: bool

    PLUGIN_PRESETS: Mapping[int, Callable[[], ProcessorPluginPreset[ProcessorConfigT]]] = {}

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ):
        super().__init__(callback=callback, generation=generation, key=key)
        self._processor_instance = None
        self._log: BackendLogger = BackendLogger.getLogger(__name__)
        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.session_config = a121.SessionConfig.from_json(file["session_config"][()])
        self.shared_state.processor_config = self.get_processor_config_cls().from_json(
            file["processor_config"][()]
        )

    def _sync_sensor_ids(self) -> None:
        if self.client is None:
            return

        available_sensor_ids = self.client.server_info.connected_sensors

        if len(available_sensor_ids) == 0:
            return

        original_session_config = self.shared_state.session_config

        adjusted_groups = core_utils.create_extended_structure(
            (
                group_idx,
                sensor_id
                if (sensor_id in available_sensor_ids)
                else next(fallback_sensor_ids_iter),
                sensor_config,
            )
            for (group_idx, sensor_id, sensor_config), fallback_sensor_ids_iter in zip(
                core_utils.iterate_extended_structure(original_session_config.groups),
                itertools.tee(available_sensor_ids, len(available_sensor_ids)),
            )
        )
        self.shared_state.session_config = a121.SessionConfig(
            adjusted_groups,
            update_rate=original_session_config.update_rate,
            extended=original_session_config.extended,
        )

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(
            file, "session_config", self.shared_state.session_config.to_json()
        )
        _create_h5_string_dataset(
            file, "processor_config", self.shared_state.processor_config.to_json()
        )

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        self.shared_state.session_config = record.session_config
        try:
            algo_group = record.get_algo_group(self.key)  # noqa: F841
        except KeyError:
            self.shared_state.processor_config = self.get_processor_config_cls()()
            self.callback(
                StatusMessage(
                    status="Could not load algo group. "
                    + "Falling back to default processor config"
                )
            )
        else:
            self.shared_state.processor_config = self.get_processor_config_cls().from_json(
                algo_group["processor_config"][()]
            )

    @is_task
    def update_session_config(self, *, session_config: a121.SessionConfig) -> None:
        self.shared_state.session_config = session_config
        self.broadcast()

    @is_task
    def update_processor_config(self, *, processor_config: ProcessorConfigT) -> None:
        self.shared_state.processor_config = processor_config
        self.broadcast()

    def _validate_session_config(self, session_config: a121.SessionConfig) -> None:
        """Hook for validating session configs."""
        pass

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        session_config = self.shared_state.session_config
        self._validate_session_config(session_config)

        assert self.client
        if recorder:
            algo_group = recorder.require_algo_group(self.key)  # noqa: F841

            # TODO: break out saving (?)
            _create_h5_string_dataset(
                algo_group, "processor_config", self.shared_state.processor_config.to_json()
            )

            self.client.attach_recorder(recorder)

        metadata = self.client.setup_session(session_config)
        self.shared_state.metadata = metadata

        self._processor_instance = self.get_processor(self.shared_state)

        self.client.start_session()

        self.callback(
            SetupMessage(
                metadata=metadata,
                session_config=session_config,
                processor_config=self.shared_state.processor_config,
            )
        )

    def end_session(self) -> None:
        if self.client is None:
            msg = "Client is not attached. Can not 'stop'."
            raise RuntimeError(msg)

        self.client.stop_session()

        recorder = self.client.detach_recorder()
        if recorder is not None:
            recorder.close()

    @classmethod
    @abc.abstractmethod
    def get_processor(
        cls, state: ProcessorBackendPluginSharedState[ProcessorConfigT]
    ) -> GenericProcessorBase[InputT, ResultT]:
        pass

    @classmethod
    @abc.abstractmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfigT]:
        pass

    @classmethod
    @abc.abstractmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        pass


class ProcessorBackendPluginBase(
    GenericProcessorBackendPluginBase[a121.Result, ProcessorConfigT, ResultT, a121.Metadata]
):
    def _validate_session_config(self, session_config: a121.SessionConfig) -> None:
        if session_config.extended:
            msg = "Extended session configs are not supported."
            raise ValueError(msg)

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = ProcessorBackendPluginSharedState(
            session_config=a121.SessionConfig(self.get_default_sensor_config()),
            processor_config=self.get_processor_config_cls()(),
        )

        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        processor_preset = preset_config()
        self.shared_state.session_config = processor_preset.session_config
        self.shared_state.processor_config = processor_preset.processor_config
        self.broadcast()

    def get_next(self) -> None:
        if self._processor_instance is None:
            msg = "Processor is None. 'start' needs to be called before 'get_next'"
            raise RuntimeError(msg)

        assert self.client
        result = self.client.get_next()
        assert isinstance(result, a121.Result)  # TODO: fix

        processor_result = self._processor_instance.process(result)

        self.callback(PlotMessage(result=processor_result))


class ExtendedProcessorBackendPluginBase(
    GenericProcessorBackendPluginBase[
        List[Dict[int, a121.Result]], ProcessorConfigT, ResultT, List[Dict[int, a121.Metadata]]
    ]
):
    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = ProcessorBackendPluginSharedState(
            session_config=a121.SessionConfig(self.get_default_sensor_config()),
            processor_config=self.get_processor_config_cls()(),
        )

        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        processor_preset = preset_config()
        self.shared_state.session_config = processor_preset.session_config
        self.shared_state.processor_config = processor_preset.processor_config
        self.broadcast()

    def get_next(self) -> None:
        if self._processor_instance is None:
            msg = "Processor is None. 'start' needs to be called before 'get_next'"
            raise RuntimeError(msg)

        assert self.client
        result = self.client.get_next()
        if isinstance(result, a121.Result):
            results = [{self.client.session_config.sensor_id: result}]
        else:
            results = result

        processor_result = self._processor_instance.process(results)

        self.callback(PlotMessage(result=processor_result))



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
#
from .result_message import EmptyResultMessage, ResultMessage
from .sensor_info_response import SensorInfoResponse
from .setup_response import SetupResponse



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool.a121._core.entities.configs.config_enums import IdleState, Profile
from acconeer.exptool.a121._core.utils import is_divisor_of, is_multiple_of
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoConfigBase,
    Controller,
    select_prf,
)

from ._processor import Processor, ProcessorConfig, ProcessorExtraResult


SPARSE_IQ_PPC = 24


def optional_profile_converter(profile: Optional[Profile]) -> Optional[Profile]:
    if profile is None:
        return None

    return Profile(profile)


def idle_state_converter(idle_state: IdleState) -> IdleState:
    return IdleState(idle_state)


@attributes_doc
@attrs.mutable(kw_only=True)
class ExampleAppConfig(AlgoConfigBase):
    surface_distance: float = attrs.field(default=1)
    """Perpendicular distance from the water surface to the sensor in meters."""

    sensor_angle: float = attrs.field(default=45)
    """Sensor angle in degrees. 0 degrees is defined as the sensor
    facing straight down to the surface."""

    num_points: int = attrs.field(default=4)
    """Number of data points in the measurement."""

    step_length: int = attrs.field(default=12)
    """Step length in points."""

    profile: Optional[a121.Profile] = attrs.field(
        default=None, converter=optional_profile_converter
    )
    """Sets the profile. If no argument is provided, the highest possible
    profile without interference of direct leakage is used to maximize SNR."""

    frame_rate: Optional[float] = attrs.field(default=None)
    """Frame rate in Hz."""

    sweep_rate: float = attrs.field(default=3000)
    """Sweep rate in Hz."""

    sweeps_per_frame: int = attrs.field(default=128)
    """Number of sweeps per frame."""

    hwaas: int = attrs.field(default=16)
    """Number of HWAAS."""

    double_buffering: bool = attrs.field(default=True)
    """Enables double buffering."""

    continuous_sweep_mode: bool = attrs.field(default=True)
    """Enables continuous sweep mode."""

    inter_frame_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.READY, converter=idle_state_converter
    )
    """Sets the inter frame idle state."""

    inter_sweep_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.READY, converter=idle_state_converter
    )
    """Sets the inter sweep idle state."""

    time_series_length: int = attrs.field(default=512)
    """Number of sweeps in the time series."""

    psd_lp_coeff: float = attrs.field(default=0.75)
    """Filter coefficient for the exponential filter of PSD over time."""

    cfar_sensitivity: float = attrs.field(default=0.15)
    """Sensitivity of the CFAR threshold. Low sensitivity will set a high threshold."""

    cfar_guard: int = attrs.field(default=6)
    """Number of frequency bins around the point of interest that
    is omitted when calculating the CFAR threshold."""

    cfar_win: int = attrs.field(default=6)
    """Number of frequency bins next to the CFAR guard from
    which the threshold level will be calculated."""

    slow_zone: int = attrs.field(default=3)
    """Half size of the number of frequency bins that are regarded as the slow zone."""

    velocity_lp_coeff: float = attrs.field(default=0.98)
    """Filter coefficient for the exponential filter of the velocity estimate."""

    max_peak_interval_s: float = attrs.field(default=4)
    """Maximal number of seconds that is tolerated between
    peaks before the estimated velocity starts decreasing."""

    @step_length.validator
    def _validate_step_length(self, _: Any, step_length: int) -> None:
        if step_length is not None and not (
            is_divisor_of(SPARSE_IQ_PPC, step_length) or is_multiple_of(SPARSE_IQ_PPC, step_length)
        ):
            msg = f"step_length must be a divisor or multiple of {SPARSE_IQ_PPC}"
            raise ValueError(msg)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if not self.sensor_angle < 90 or not self.sensor_angle > 0:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sensor_angle",
                    "Sensor angle must be between 0 and 90 degrees",
                )
            )

        optimal_distance = self.surface_distance / np.cos(np.radians(self.sensor_angle))
        optimal_point = int(np.ceil(optimal_distance / APPROX_BASE_STEP_LENGTH_M))

        start_point = int(
            np.around(optimal_point - np.floor((self.num_points - 1) / 2) * self.step_length)
        )

        end_point = start_point + (self.num_points - 1) * self.step_length

        if end_point * APPROX_BASE_STEP_LENGTH_M > a121.PRF.PRF_5_2_MHz.mmd:
            msg = "The required measurable distance exceeds maximum measureable distance. Decrease surface distance, sensor angle, number of distance points or step length."
            validation_results.append(
                a121.ValidationError(
                    self,
                    "surface_distance",
                    msg,
                )
            )
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sensor_angle",
                    msg,
                )
            )
            validation_results.append(
                a121.ValidationError(
                    self,
                    "num_points",
                    msg,
                )
            )
            validation_results.append(
                a121.ValidationError(
                    self,
                    "step_length",
                    msg,
                )
            )

        if start_point * APPROX_BASE_STEP_LENGTH_M <= self.surface_distance:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sensor_angle",
                    (
                        "Start point must be > surface distance. Increase sensor angle, "
                        "decrease step length or decrease number of distance points"
                    ),
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ExampleAppResult:
    velocity: float = attrs.field()
    """Estimated velocity."""

    distance_m: float = attrs.field()
    """Distance in meters used for the current velocity estimate."""

    processor_extra_result: ProcessorExtraResult = attrs.field()
    service_result: a121.Result = attrs.field()


class ExampleApp(Controller[ExampleAppConfig, ExampleAppResult]):
    MIN_DIST_M = {
        a121.Profile.PROFILE_1: None,
        a121.Profile.PROFILE_2: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_2],
        a121.Profile.PROFILE_3: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_3],
        a121.Profile.PROFILE_4: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_4],
        a121.Profile.PROFILE_5: 2 * ENVELOPE_FWHM_M[a121.Profile.PROFILE_5],
    }

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        example_app_config: ExampleAppConfig,
    ) -> None:
        super().__init__(client=client, config=example_app_config)
        self.sensor_id = sensor_id

        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, _algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        sensor_config = self._get_sensor_config(self.config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        processor_config = self._get_processor_config(self.config)

        self.processor = Processor(
            sensor_config=sensor_config,
            metadata=metadata,
            processor_config=processor_config,
            context=None,
        )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("surface_velocity")
                _record_algo_data(
                    _algo_group,
                    self.sensor_id,
                    self.config,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.started = True

    @classmethod
    def _get_sensor_config(cls, config: ExampleAppConfig) -> a121.SensorConfig:
        optimal_distance = config.surface_distance / np.cos(np.radians(config.sensor_angle))
        optimal_point = int(np.ceil(optimal_distance / APPROX_BASE_STEP_LENGTH_M))

        start_point = int(
            np.around(optimal_point - np.floor((config.num_points - 1) / 2) * config.step_length)
        )
        end_point = start_point + (config.num_points - 1) * config.step_length

        if config.profile is not None:
            profile = config.profile
        else:
            viable_profiles = [
                k
                for k, v in cls.MIN_DIST_M.items()
                if v is None or v <= start_point * APPROX_BASE_STEP_LENGTH_M
            ]
            profile = viable_profiles[-1]

        return a121.SensorConfig(
            profile=profile,
            start_point=start_point,
            num_points=config.num_points,
            step_length=config.step_length,
            prf=select_prf(end_point, profile),
            hwaas=config.hwaas,
            sweeps_per_frame=config.sweeps_per_frame,
            frame_rate=config.frame_rate,
            sweep_rate=config.sweep_rate,
            continuous_sweep_mode=config.continuous_sweep_mode,
            double_buffering=config.double_buffering,
            inter_frame_idle_state=config.inter_frame_idle_state,
            inter_sweep_idle_state=config.inter_sweep_idle_state,
        )

    @classmethod
    def _get_processor_config(cls, config: ExampleAppConfig) -> ProcessorConfig:
        return ProcessorConfig(
            surface_distance=config.surface_distance,
            time_series_length=config.time_series_length,
            slow_zone=config.slow_zone,
            psd_lp_coeff=config.psd_lp_coeff,
            cfar_guard=config.cfar_guard,
            cfar_win=config.cfar_win,
            cfar_sensitivity=config.cfar_sensitivity,
            velocity_lp_coeff=config.velocity_lp_coeff,
            max_peak_interval_s=config.max_peak_interval_s,
        )

    def get_next(self) -> ExampleAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        return ExampleAppResult(
            velocity=processor_result.estimated_v,
            distance_m=processor_result.distance_m,
            processor_extra_result=processor_result.extra_result,
            service_result=result,
        )

    def update_config(self, config: ExampleAppConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: ExampleAppConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "example_app_config", config.to_json())


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, ExampleAppConfig]:
    sensor_id = algo_group["sensor_id"][()]
    config = ExampleAppConfig.from_json(algo_group["example_app_config"][()])
    return sensor_id, config



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import typing as t
from time import time

import h5py
import typing_extensions as te


_ConfigT = t.TypeVar("_ConfigT", contravariant=True)
_MetadataT = t.TypeVar("_MetadataT", contravariant=True)
_ResultT = t.TypeVar("_ResultT", contravariant=True)
_ServerInfoT = t.TypeVar("_ServerInfoT", contravariant=True)


class H5Saver(te.Protocol[_ConfigT, _MetadataT, _ResultT, _ServerInfoT]):
    """Interface for savers handling writing results to file"""

    def _start(self) -> None: ...

    def _write_server_info(self, group: h5py.Group, server_info: _ServerInfoT) -> None: ...

    def _start_session(
        self, group: h5py.Group, *, config: _ConfigT, metadata: _MetadataT, **kwargs: t.Any
    ) -> None: ...

    def _sample(self, group: h5py.Group, results: t.Iterable[_ResultT]) -> None: ...

    def _stop_session(self, group: h5py.Group) -> None: ...


class ChunkedH5Saver(H5Saver[_ConfigT, _MetadataT, _ResultT, _ServerInfoT]):
    """Chunks result before writing to file to decrease file system access

    :param saver:
        Saver used to write results to file
    :param _chunk_size:
        If given, data will be written to file every ``_chunk_size`` samples.

        If not given, data will be written at least every 512:th sample, or at least once per
        second, whichever comes first.

        Setting a small chunk size (e.g. 1) may degrade performance for high sample rates.

        Internal parameter, subject to change.
    """

    _AUTO_CHUNK_MAX_SIZE = 512
    _AUTO_CHUNK_MAX_TIME = 1.0

    _saver: H5Saver[_ConfigT, _MetadataT, _ResultT, _ServerInfoT]
    _chunk_size: t.Optional[int]
    _chunk_buffer: list[_ResultT]
    _last_write_time: float

    def __init__(
        self,
        saver: H5Saver[_ConfigT, _MetadataT, _ResultT, _ServerInfoT],
        _chunk_size: t.Optional[int] = None,
    ) -> None:
        self._saver = saver
        self._chunk_size = _chunk_size
        self._chunk_buffer = []
        self._last_write_time = 0.0

    def _start(self) -> None:
        self._saver._start()
        self._last_write_time = time()

    def _write_server_info(self, group: h5py.Group, server_info: _ServerInfoT) -> None:
        self._saver._write_server_info(group, server_info)

    def _start_session(
        self, group: h5py.Group, *, config: _ConfigT, metadata: _MetadataT, **kwargs: t.Any
    ) -> None:
        self._saver._start_session(group, config=config, metadata=metadata, **kwargs)
        self._last_write_time = time()

    def _sample(self, group: h5py.Group, results: t.Iterable[_ResultT]) -> None:
        for result in results:
            self._chunk_buffer.append(result)

        if self._chunk_size is None:
            reached_size_limit = len(self._chunk_buffer) >= self._AUTO_CHUNK_MAX_SIZE
            reached_time_limit = (time() - self._last_write_time) >= self._AUTO_CHUNK_MAX_TIME
            write = reached_size_limit or reached_time_limit
        else:
            write = len(self._chunk_buffer) == self._chunk_size

        if write:
            self._saver._sample(group, self._chunk_buffer)
            self._chunk_buffer = []
            self._last_write_time = time()

    def _stop_session(self, group: h5py.Group) -> None:
        if len(self._chunk_buffer) > 0:
            self._saver._sample(group, self._chunk_buffer)
            self._chunk_buffer = []
        self._saver._stop_session(group)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import typing as t

import typing_extensions as te

from acconeer.exptool._core import recording
from acconeer.exptool.a121._core.entities import (
    Metadata,
    Result,
    SensorCalibration,
    ServerInfo,
    SessionConfig,
)


class Recorder(
    recording.Recorder[
        SessionConfig,  # Config type
        t.List[t.Dict[int, Metadata]],  # Metadata type
        t.List[t.Dict[int, Result]],  # Result type
        ServerInfo,  # Server info type
    ],
    te.Protocol,
):
    def _start_session(
        self,
        *,
        config: SessionConfig,
        metadata: t.List[t.Dict[int, Metadata]],
        calibrations: t.Optional[t.Dict[int, SensorCalibration]] = None,
        calibrations_provided: t.Optional[t.Dict[int, bool]] = None,
    ) -> None: ...


RecorderAttachable: te.TypeAlias = recording.RecorderAttachable[Recorder]



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from . import builtin_persistors, core, optimizing_persistors
from .api import (
    deserialize,
    register_json_presentable,
    register_persistor,
    serialize,
    try_deserialize,
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool import type_migration as tm
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoConfigBase,
    Controller,
)
from acconeer.exptool.a121.algo.breathing._processor import (
    AppState,
    BreathingProcessorConfig,
    BreathingProcessorResult,
    Processor,
    ProcessorConfig,
)
from acconeer.exptool.a121.algo.presence import ProcessorConfig as PresenceProcessorConfig
from acconeer.exptool.a121.algo.presence import ProcessorResult as PresenceProcessorResult
from acconeer.exptool.a121.algo.presence._processors import processor_config_timeline


def get_presence_config() -> PresenceProcessorConfig:
    presence_config = PresenceProcessorConfig()
    presence_config.intra_detection_threshold = 4.0
    return presence_config


@attributes_doc
@attrs.mutable(kw_only=True)
class RefAppConfig(AlgoConfigBase):
    use_presence_processor: bool = attrs.field(default=True)
    """If True, use the presence processor to determine distance to subject."""

    num_distances_to_analyze: int = attrs.field(default=3)
    """Indicates the number of distance to analyzed, centered around the distance where presence
    is detected."""

    distance_determination_duration: float = attrs.field(default=5.0)
    """Time for the presence processor to determine distance to presence."""

    start_m: float = attrs.field(default=0.3)
    """Start of measurement range (m)."""

    end_m: float = attrs.field(default=1.5)
    """End of measurement range (m)."""

    hwaas: int = attrs.field(default=32)
    """HWAAS."""

    profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_3, converter=a121.Profile)
    """Profile."""

    frame_rate: float = attrs.field(default=20.0)
    """Frame rate."""

    sweeps_per_frame: int = attrs.field(default=16)
    """Sweeps per frame."""

    breathing_config: BreathingProcessorConfig = attrs.field(factory=BreathingProcessorConfig)
    """Breathing configuration."""

    presence_config: PresenceProcessorConfig = attrs.field(factory=get_presence_config)
    """Presence configuration."""

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> RefAppConfig:
        presence_dict = d["presence_config"]
        if presence_dict is not None:
            try:
                d["presence_config"] = processor_config_timeline.migrate(presence_dict)
            except tm.core.MigrationErrorGroup as exc:
                raise TypeError() from exc

        breathing_dict = d["breathing_config"]
        if breathing_dict is not None:
            d["breathing_config"] = BreathingProcessorConfig.from_dict(breathing_dict)
        return cls(**d)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.end_m < self.start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    "End point must be greater than start point",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class RefAppResult:
    app_state: AppState
    """Application state."""

    distances_being_analyzed: Optional[Tuple[int, int]] = None
    """Range where breathing is being analyzed."""

    presence_result: PresenceProcessorResult
    """Presence processor result."""

    breathing_result: Optional[BreathingProcessorResult] = attrs.field(default=None)
    """Breathing processor result."""


class RefApp(Controller[RefAppConfig, RefAppResult]):
    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        ref_app_config: RefAppConfig,
    ) -> None:
        super().__init__(client=client, config=ref_app_config)

        self.sensor_config = get_sensor_config(ref_app_config)

        self.processor_config = ProcessorConfig()
        self.processor_config.num_distances_to_analyze = ref_app_config.num_distances_to_analyze
        self.processor_config.distance_determination_duration = (
            ref_app_config.distance_determination_duration
        )
        self.processor_config.use_presence_processor = ref_app_config.use_presence_processor
        self.processor_config.presence_config = ref_app_config.presence_config
        self.processor_config.breathing_config = ref_app_config.breathing_config

        self.sensor_id = sensor_id

        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        self.metadata = self.client.setup_session(
            a121.SessionConfig({self.sensor_id: self.sensor_config})
        )

        assert not isinstance(self.metadata, list)

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                algo_group = recorder.require_algo_group("breathing")
                _record_algo_data(
                    algo_group,
                    self.sensor_id,
                    self.config,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

            self.client.attach_recorder(recorder)

        self.processor = Processor(
            sensor_config=self.sensor_config,
            processor_config=self.processor_config,
            metadata=self.metadata,
        )

        self.client.start_session()

        self.started = True

    def get_next(self) -> RefAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()
        assert not isinstance(result, list)

        processor_result = self.processor.process(result)

        ref_app_result = RefAppResult(
            app_state=processor_result.app_state,
            distances_being_analyzed=processor_result.distances_being_analyzed,
            breathing_result=processor_result.breathing_result,
            presence_result=processor_result.presence_result,
        )

        return ref_app_result

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    ref_app_config: RefAppConfig,
) -> None:
    algo_group.create_dataset(
        "ref_app_sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "ref_app_config", ref_app_config.to_json())


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, RefAppConfig]:
    sensor_id = int(algo_group["ref_app_sensor_id"][()])
    config = RefAppConfig.from_json(algo_group["ref_app_config"][()])
    return sensor_id, config


def _determine_step_length(ref_app_config: RefAppConfig) -> int:
    # Determine step length based on selected profile. The target length is half FWHM.
    VALID_STEP_LENGTHS = [1, 2, 4, 6, 8, 12, 24]
    step_length = int(
        ENVELOPE_FWHM_M[a121.Profile(ref_app_config.profile)] / APPROX_BASE_STEP_LENGTH_M / 2
    )

    if step_length < VALID_STEP_LENGTHS[-1]:
        idx_closest = np.sum(np.array(VALID_STEP_LENGTHS) <= step_length) - 1
        step_length = int(VALID_STEP_LENGTHS[idx_closest])
    else:
        step_length = int((step_length // VALID_STEP_LENGTHS[-1]) * VALID_STEP_LENGTHS[-1])

    return step_length


def get_sensor_config(ref_app_config: RefAppConfig) -> a121.SensorConfig:
    """Defines sensor configuration based on reference application configuration."""
    step_length = _determine_step_length(ref_app_config)

    start_point = int(ref_app_config.start_m // APPROX_BASE_STEP_LENGTH_M)
    num_points = int(
        np.ceil(
            (ref_app_config.end_m - ref_app_config.start_m)
            / (step_length * APPROX_BASE_STEP_LENGTH_M)
        )
        + 1
    )

    c = a121.SensorConfig()
    c.start_point = start_point
    c.num_points = num_points
    c.step_length = step_length
    c.hwaas = ref_app_config.hwaas
    c.profile = ref_app_config.profile
    c.sweeps_per_frame = ref_app_config.sweeps_per_frame
    c.frame_rate = ref_app_config.frame_rate

    return c



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import math
import statistics
import typing as t
from collections import deque

import attrs
import typing_extensions as te


@attrs.frozen
class _RateStats:
    rate: float
    rate_warning: bool
    jitter: float
    jitter_warning: bool

    @classmethod
    def invalid(cls) -> te.Self:
        return cls(rate=math.nan, rate_warning=False, jitter=math.nan, jitter_warning=False)


@attrs.mutable
class _RateCalculator:
    """Stateful class that monitors result rate and jitter.

    Since the ``_RateCalculator`` compares time differances between results, the first call
    to ``update`` will "prime" it and will not produce any meaningful statistics.

    >>> rc = _RateCalculator(ticks_per_second=1000, tick_period=100)
    >>> rc.update(tick=0, frame_delayed=False)
    _RateStats(rate=nan, rate_warning=False, jitter=nan, jitter_warning=False)

    Once the ``_RateCalculator`` is passed its second result, statistics are meaningful:

    >>> rc.update(tick=100, frame_delayed=False)
    _RateStats(rate=10.0, rate_warning=False, jitter=0.0, jitter_warning=False)

    Once the passed result aren't exactly equidistant in time, jitter will be non-zero:

    >>> rc.update(tick=201, frame_delayed=False)
    _RateStats(rate=9.95..., rate_warning=False, jitter=0.0005..., jitter_warning=False)

    If the result has the indication ``frame_delayed``, ``rate_warning`` will always be true.

    >>> rc.update(tick=300, frame_delayed=True)
    _RateStats(rate=10.0, rate_warning=True, jitter=0.0008..., jitter_warning=False)

    If the measured rate or its jitter becomes too large, both warning flag will be
    set to True:

    >>> rc.update(tick=600, frame_delayed=False)
    _RateStats(rate=6.66..., rate_warning=True, jitter=0.08..., jitter_warning=True)
    """

    _JITTER_WARNING_LIMIT: t.ClassVar[float] = 1.0e-3  # Based on testing

    ticks_per_second: int
    tick_period: int

    ticks: deque[int] = attrs.field(factory=lambda: deque([], maxlen=200), init=False)

    @property
    def tick_period_upper_bound(self) -> float:
        if self.tick_period == 0:
            return math.inf
        else:
            return self.tick_period * 1.01 + 10

    @property
    def tick_diffs(self) -> list[int]:
        return [rhs - lhs for lhs, rhs in zip(self.ticks, list(self.ticks)[1:])]

    def update(self, tick: int, frame_delayed: bool) -> _RateStats:
        self.ticks.append(tick)

        if len(self.ticks) < 2:
            return _RateStats.invalid()

        measured_tick_period = statistics.mean(self.tick_diffs)
        measured_rate = 1.0 / (measured_tick_period / self.ticks_per_second)
        rate_warning = frame_delayed or measured_tick_period > self.tick_period_upper_bound

        if len(self.tick_diffs) < 2:
            return _RateStats(measured_rate, rate_warning, jitter=0.0, jitter_warning=False)
        else:
            jitter_s = statistics.pstdev(self.tick_diffs) / self.ticks_per_second
            jitter_warning = jitter_s > self._JITTER_WARNING_LIMIT
            return _RateStats(measured_rate, rate_warning, jitter_s, jitter_warning)



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from ._enums import (
    ConnectionInterface,
    ConnectionState,
    PluginFamily,
    PluginGeneration,
    PluginState,
)
from ._exceptions import HandledException
from ._version_checker import check_package_outdated, get_latest_changelog
from .app import main
from .app_model import AppModel
from .backend import (
    ApplicationClient,
    BackendLogger,
    BackendPlugin,
    GeneralMessage,
    Message,
    PluginStateMessage,
    is_task,
)
from .plugin_loader import register_plugin
from .pluginbase import (
    PgPlotPlugin,
    PlotPluginBase,
    PluginPresetBase,
    PluginSpecBase,
    ViewPluginBase,
    visual_policies,
)
from .storage import get_temp_dir, get_temp_h5_path
from .ui import (
    AttrsConfigEditor,
    GroupBox,
    MiscErrorView,
    PidgetFactoryMapping,
    PidgetGroupFactoryMapping,
    TwoSensorIdsEditor,
    icons,
    pidgets,
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import contextlib
import logging
from enum import Enum, auto
from typing import Any, Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtGui import QFont
from PySide6.QtWidgets import QLabel, QPushButton, QTabWidget, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121, opser
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    HandledException,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    PluginStateMessage,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import (
    CollapsibleWidget,
    GotoResourceTabButton,
)
from acconeer.exptool.app.new.ui.components.a121 import SensorConfigEditor
from acconeer.exptool.app.new.ui.components.json_save_load_buttons import (
    JsonButtonOperations,
    PresentationType,
)

from . import _pidget_mapping
from ._configs import get_high_accuracy_detector_config
from ._context import detector_context_timeline
from ._detector import (
    DetailedStatus,
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
    _load_algo_data,
)
from ._translation import detector_config_to_session_config


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_ids: list[int] = attrs.field(factory=lambda: [1])
    config: DetectorConfig = attrs.field(factory=DetectorConfig)
    context: DetectorContext = attrs.field(factory=DetectorContext)


class PluginPresetId(Enum):
    BALANCED = auto()
    HIGH_ACCURACY = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    num_curves: int
    start_m: float
    end_m: float
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], DetectorConfig]] = {
        PluginPresetId.BALANCED.value: lambda: DetectorConfig(),
        PluginPresetId.HIGH_ACCURACY.value: lambda: get_high_accuracy_detector_config(),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)
        self._detector_instance: Optional[Detector] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            # Try to use the sensor ids from the last calibration
            if self.shared_state.context.sensor_ids:
                self.shared_state.sensor_ids = self.shared_state.context.sensor_ids

            for i in range(len(self.shared_state.sensor_ids)):
                if len(sensor_ids) > 0 and self.shared_state.sensor_ids[i] not in sensor_ids:
                    self.shared_state.sensor_ids[i] = sensor_ids[0]

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_sensor_ids(self, *, sensor_ids: list[int]) -> None:
        self.shared_state.sensor_ids = sensor_ids
        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        context_group = file.create_group("context")
        opser.serialize(self.shared_state.context, context_group)

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])
        self.shared_state.context = detector_context_timeline.migrate(file["context"])

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.context = context
        self.shared_state.sensor_ids = list(next(iter(record.session_config.groups)).keys())

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._detector_instance = Detector(
            client=self.client,
            sensor_ids=self.shared_state.sensor_ids,
            detector_config=self.shared_state.config,
            context=self.shared_state.context,
        )
        self._detector_instance.start(recorder)
        self.callback(
            SetupMessage(
                num_curves=len(self._detector_instance.processor_specs),
                start_m=self.shared_state.config.start_m,
                end_m=self.shared_state.config.end_m,
            )
        )

    def end_session(self) -> None:
        assert self._detector_instance
        if self._recorder is not None:
            self._recorder.close()
        self._detector_instance.stop()

    def get_next(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError

        assert self.client
        result = self._detector_instance.get_next()

        self.callback(backend.PlotMessage(result=result))

    @is_task
    def calibrate_detector(self) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError

        if not self.client.connected:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

        try:
            self._detector_instance = Detector(
                client=self.client,
                sensor_ids=self.shared_state.sensor_ids,
                detector_config=self.shared_state.config,
                context=None,
            )
            self._detector_instance.calibrate_detector()
        except Exception as exc:
            msg = "Failed to calibrate detector"
            raise HandledException(msg) from exc
        finally:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        self.shared_state.context = self._detector_instance.context
        self.broadcast()


class PlotPlugin(PgPlotPlugin):
    _DISTANCE_HISTORY_SPAN_MARGIN = 0.05
    _DISTANCE_HISTORY_LEN = 100
    _MAX_NUM_MINOR_PEAKS = 4

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[dict[int, DetectorResult]] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(message.num_curves, message.start_m, message.end_m)
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(multi_sensor_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(self, num_curves: int, start_m: float, end_m: float) -> None:
        self.plot_layout.clear()

        self.num_curves = num_curves
        self.start_m = start_m
        self.end_m = end_m

        self.main_peak_history = np.full(self._DISTANCE_HISTORY_LEN, fill_value=np.nan)
        self.minor_peaks_history = np.full(
            (self._MAX_NUM_MINOR_PEAKS, self._DISTANCE_HISTORY_LEN), fill_value=np.nan
        )

        win = self.plot_layout

        # Sweep plot
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.sweep_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.threshold_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweep_curves[0], "Sweep")
        sweep_plot_legend.addItem(self.threshold_curves[0], "Threshold")

        font = QFont()
        font.setPixelSize(16)
        self.sweep_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
            color=pg.mkColor(0xFF, 0xFF, 0xFF, 200),
        )
        self.sweep_text_item.setFont(font)
        self.sweep_text_item.hide()
        self.sweep_plot.addItem(self.sweep_text_item)

        self.sweep_main_peak_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5, dash=[2, 8]))
        self.sweep_main_peak_line.hide()
        self.sweep_plot.addItem(self.sweep_main_peak_line)

        # History plot
        self.dist_history_plot = win.addPlot(row=1, col=0)
        self.dist_history_plot.setMenuEnabled(False)
        self.dist_history_plot.showGrid(x=True, y=True)
        self.dist_history_plot.addLegend()
        self.dist_history_plot.setLabel("left", "Estimated distance (m)")
        self.dist_history_plot.addItem(pg.PlotDataItem())
        self.dist_history_plot.setXRange(0, self._DISTANCE_HISTORY_LEN)

        brush = et.utils.pg_brush_cycler(0)
        symbol_kw_main = dict(
            symbol="o", symbolSize=5, symbolBrush=brush, symbolPen=None, pen=None
        )
        feat_kw = dict(**symbol_kw_main)
        self.main_peak_history_curve = self.dist_history_plot.plot(**feat_kw)

        minor_colors = [et.utils.color_cycler(i) for i in range(1, self._MAX_NUM_MINOR_PEAKS)]
        self.minor_peaks_history_curves = [
            self.dist_history_plot.plot(
                **dict(
                    **dict(
                        symbol="o",
                        symbolSize=5,
                        symbolBrush=pg.mkBrush(color),
                        symbolPen=None,
                        pen=None,
                    ),
                )
            )
            for color in minor_colors
        ]

        # History legend
        self.history_plot_legend = pg.LegendItem()
        self.dist_history_plot.addItem(self.history_plot_legend)

        # Smoothing
        self.sweep_smooth_max = et.utils.SmoothMax()
        self.distance_hist_smooth_lim = et.utils.SmoothLimits(tau_decay=0.5)

    def draw_plot_job(self, *, multi_sensor_result: dict[int, DetectorResult]) -> None:
        # Get the first element as the plugin only supports single sensor operation.
        (result,) = list(multi_sensor_result.values())

        assert result.distances is not None
        assert result.strengths is not None

        strengths = result.strengths
        distances = result.distances

        # Update main peak history
        self.main_peak_history = np.roll(self.main_peak_history, shift=-1)
        if len(distances) != 0:
            self.main_peak_history[-1] = distances[0]
        else:
            self.main_peak_history[-1] = np.nan

        # Update minor peaks history
        num_minor_peaks = min(len(distances) - 1, self._MAX_NUM_MINOR_PEAKS)
        new_history = np.full(self._MAX_NUM_MINOR_PEAKS, fill_value=np.nan)
        self.minor_peaks_history = np.roll(self.minor_peaks_history, shift=-1)
        if num_minor_peaks > 0:
            new_history[:num_minor_peaks] = np.array(distances[1 : 1 + num_minor_peaks])
        self.minor_peaks_history[:, -1] = new_history

        # Sweep plot
        max_val_in_plot = 0
        for idx, processor_result in enumerate(result.processor_results):
            assert processor_result.extra_result.used_threshold is not None
            assert processor_result.extra_result.distances_m is not None
            assert processor_result.extra_result.abs_sweep is not None

            abs_sweep = processor_result.extra_result.abs_sweep
            threshold = processor_result.extra_result.used_threshold
            distances_m = processor_result.extra_result.distances_m

            self.sweep_curves[idx].setData(distances_m, abs_sweep)
            self.threshold_curves[idx].setData(distances_m, threshold)

            max_val_in_subsweep = max(max(threshold), max(abs_sweep))
            if max_val_in_plot < max_val_in_subsweep:
                max_val_in_plot = max_val_in_subsweep

        self.sweep_plot.setYRange(0, self.sweep_smooth_max.update(max_val_in_plot))

        if len(distances) != 0:
            text_y_pos = self.sweep_plot.getAxis("left").range[1] * 0.95
            text_x_pos = (
                self.sweep_plot.getAxis("bottom").range[1]
                + self.sweep_plot.getAxis("bottom").range[0]
            ) / 2.0
            self.sweep_text_item.setPos(text_x_pos, text_y_pos)
            self.sweep_text_item.setHtml("Main peak distance: {:.3f} m".format(distances[0]))
            self.sweep_text_item.show()

            self.sweep_main_peak_line.setPos(distances[0])
            self.sweep_main_peak_line.show()
        else:
            self.sweep_text_item.hide()
            self.sweep_main_peak_line.hide()

        # History plot
        self.history_plot_legend.clear()

        if np.any(~np.isnan(self.main_peak_history)):
            self.main_peak_history_curve.setData(self.main_peak_history)
        else:
            self.main_peak_history_curve.setData([0])

        if strengths.size > 0:
            string_to_display = "Main peak strength : {:.1f} dB".format(strengths[0])
            self.history_plot_legend.addItem(self.main_peak_history_curve, string_to_display)

        for sec_peak_idx, curve in enumerate(self.minor_peaks_history_curves):
            if np.any(~np.isnan(self.minor_peaks_history[sec_peak_idx, :])):
                curve.setData(self.minor_peaks_history[sec_peak_idx, :])
            else:
                curve.setData([0])

            if sec_peak_idx + 1 <= num_minor_peaks:
                string_to_display = "Minor peak strength : {:.1f} dB".format(
                    strengths[sec_peak_idx + 1]
                )
                self.history_plot_legend.addItem(curve, string_to_display)

        # Update limits of the history plot
        max_vals = []
        min_vals = []
        if np.any(~np.isnan(self.main_peak_history)):
            max_vals.append(np.nanmax(self.main_peak_history))
            min_vals.append(np.nanmin(self.main_peak_history))

        if np.any(~np.isnan(self.minor_peaks_history)):
            max_vals.append(np.nanmax(self.minor_peaks_history))
            min_vals.append(np.nanmin(self.minor_peaks_history))

        if min_vals:
            lims = self.distance_hist_smooth_lim.update([min(min_vals), max(max_vals)])
        else:
            lims = self.distance_hist_smooth_lim.update([self.start_m, self.end_m])
        lower_lim = max(0.0, lims[0] - self._DISTANCE_HISTORY_SPAN_MARGIN)
        upper_lim = lims[1] + self._DISTANCE_HISTORY_SPAN_MARGIN
        self.dist_history_plot.setYRange(lower_lim, upper_lim)

        # Update position of legend
        y_range = self.dist_history_plot.getAxis("left").range
        x_range = self.dist_history_plot.getAxis("bottom").range
        legend_y_pos = (y_range[1] - y_range[0]) * 0.95 + y_range[0]
        legend_x_pos = (x_range[1]) * 0.01
        self.history_plot_legend.setPos(legend_x_pos, legend_y_pos)


class ViewPlugin(A121ViewPluginBase):
    sensor_config_editors: list[SensorConfigEditor]

    TEXT_MSG_MAP = {
        DetailedStatus.OK: "Ready to start.",
        DetailedStatus.END_LESSER_THAN_START: "'Range end' point must be greater than 'Range "
        + "start'.",
        DetailedStatus.CONTEXT_MISSING: "Run detector calibration.",
        DetailedStatus.CALIBRATION_MISSING: "Run detector calibration.",
        DetailedStatus.CONFIG_MISMATCH: (
            "Current configuration does not match the configuration "
            + "used during detector calibration. Run detector calibration."
        ),
    }

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.calibrate_detector_button = QPushButton(icons.CALIBRATE(), "Calibrate detector")
        self.calibrate_detector_button.clicked.connect(self._on_calibrate_detector)

        self.defaults_button = QPushButton(icons.RESTORE(), "Reset settings and calibrations")
        self.defaults_button.clicked.connect(self._send_defaults_request)

        self.message_box = QLabel(self.sticky_widget)
        self.message_box.setWordWrap(True)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.calibrate_detector_button, 1, 0, 1, -1)
        button_group.layout().addWidget(self.defaults_button, 3, 0, 1, -1)
        button_group.layout().addWidget(self.message_box, 4, 0, 1, -1)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)

        self.config_editor = AttrsConfigEditor(
            title="Detector parameters",
            factory_mapping=self.get_pidget_mapping(),
            config_type=DetectorConfig,
            extra_presenter=_detector_config_presenter,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)

        goto_resource_tab_button = GotoResourceTabButton()
        goto_resource_tab_button.clicked.connect(
            lambda: app_model.sig_resource_tab_input_block_requested.emit(
                self.config_editor.get_data()
            )
        )

        self.sensor_config_editor_tabs = QTabWidget()
        self.sensor_config_editor_tabs.setStyleSheet("QTabWidget::pane { padding: 5px;}")
        self.sensor_config_editors = []
        self.range_labels = ["Close range", "Far range"]

        for label in self.range_labels:
            sensor_config_editor = SensorConfigEditor(
                json_button_operations=JsonButtonOperations.SAVE
            )
            sensor_config_editor.set_read_only(True)
            self.sensor_config_editors.append(sensor_config_editor)
            self.sensor_config_editor_tabs.addTab(sensor_config_editor, label)

        self.collapsible_widget = CollapsibleWidget(
            "Sensor config info", self.sensor_config_editor_tabs, self.scrolly_widget
        )

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)

        sticky_layout.addWidget(button_group)

        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        scrolly_layout.addWidget(goto_resource_tab_button)
        scrolly_layout.addWidget(self.misc_error_view)
        scrolly_layout.addWidget(sensor_selection_group)
        scrolly_layout.addWidget(self.config_editor)
        scrolly_layout.addWidget(self.collapsible_widget)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return _pidget_mapping.get_pidget_mapping()

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.message_box.setText("")
        else:
            (sensor_id,) = state.sensor_ids
            self.sensor_id_pidget.set_data(sensor_id)
            self.config_editor.set_data(state.config)

            self._update_sensor_configs_view(state.config, [sensor_id])

            detector_status = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            )

            self.message_box.setText(self.TEXT_MSG_MAP[detector_status.detector_state])

            results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

    def _update_sensor_configs_view(self, config: DetectorConfig, sensor_ids: list[int]) -> None:
        with contextlib.suppress(Exception):
            session_config = detector_config_to_session_config(config, sensor_ids)
            tab_visible = [False, False]
            for group in session_config.groups:
                for _, sensor_config in group.items():
                    index = 0 if sensor_config.subsweeps[0].enable_loopback else 1
                    tab_visible[index] = True
                    sensor_config_editor = self.sensor_config_editors[index]
                    sensor_config_editor.set_data(sensor_config)

            for i, sensor_config_editor in enumerate(self.sensor_config_editors):
                index = self.sensor_config_editor_tabs.indexOf(sensor_config_editor)
                self.sensor_config_editor_tabs.setTabVisible(index, tab_visible[i])

            self.collapsible_widget.widget = self.sensor_config_editor_tabs

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[self.defaults_button, self.config_editor, self.sensor_id_pidget],
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

        state = app_model.backend_plugin_state

        if state is None:
            detector_ready = False
            config_valid = False
        else:
            detector_ready = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            ).ready_to_start

            config_valid = self._config_valid(state.config) and self.config_editor.is_ready

        self.calibrate_detector_button.setEnabled(
            visual_policies.start_button_enabled(app_model, extra_condition=config_valid),
        )
        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=detector_ready and config_valid
            )
        )

    def _config_valid(self, config: Optional[DetectorConfig]) -> bool:
        if config is None:
            return False

        try:
            config.validate()
        except a121.ValidationError:
            return False
        else:
            return True

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_ids.rpc(self.app_model.put_task, sensor_ids=[sensor_id])

    def _on_config_update(self, config: DetectorConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_calibrate_detector(self) -> None:
        BackendPlugin.calibrate_detector.rpc(self.app_model.put_task)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)


def _detector_config_presenter(config: Any, presentation_type: PresentationType) -> Optional[str]:
    if isinstance(config, DetectorConfig) and presentation_type is PresentationType.C_SET_CONFIG:
        max_step_length = 0 if config.max_step_length is None else config.max_step_length

        return f"""
static void set_config(acc_detector_distance_config_t *config, distance_preset_config_t preset)
{{
    // This snippet is generated to be compatible with RSS A121 v1.0.0
    // If there is a version missmatch the snippet might need some modification

    (void)preset;

    acc_detector_distance_config_sensor_set(config, SENSOR_ID);

    acc_detector_distance_config_start_set(config, {config.start_m:.3f}f);
    acc_detector_distance_config_end_set(config, {config.end_m:.3f}f);
    acc_detector_distance_config_max_step_length_set(config, {max_step_length}U);

    acc_detector_distance_config_signal_quality_set(config, {float(config.signal_quality):.3f}f);
    acc_detector_distance_config_max_profile_set(config, ACC_CONFIG_{config.max_profile.name});
    acc_detector_distance_config_peak_sorting_set(config, ACC_DETECTOR_DISTANCE_PEAK_SORTING_{config.peaksorting_method.name});
    acc_detector_distance_config_reflector_shape_set(config, ACC_DETECTOR_DISTANCE_REFLECTOR_SHAPE_{config.reflector_shape.name});

    acc_detector_distance_config_threshold_method_set(config, ACC_DETECTOR_DISTANCE_THRESHOLD_METHOD_{config.threshold_method.name});
    acc_detector_distance_config_num_frames_recorded_threshold_set(config, {config.num_frames_in_recorded_threshold}U);
    acc_detector_distance_config_fixed_threshold_value_set(config, {config.fixed_threshold_value:.3f}f);
    acc_detector_distance_config_threshold_sensitivity_set(config, {config.threshold_sensitivity:.3f}f);

    acc_detector_distance_config_close_range_leakage_cancellation_set(config, {str(config.close_range_leakage_cancellation).lower()});
}}
"""

    return None


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


DISTANCE_DETECTOR_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="distance_detector",
    title="Distance detector",
    docs_link="https://docs.acconeer.com/en/latest/detectors/a121/distance_detector.html",
    description="Easily measure distance to objects.",
    family=PluginFamily.DETECTOR,
    presets=[
        PluginPresetBase(name="Balanced", preset_id=PluginPresetId.BALANCED),
        PluginPresetBase(name="High accuracy", preset_id=PluginPresetId.HIGH_ACCURACY),
    ],
    default_preset_id=PluginPresetId.BALANCED,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

import acconeer.exptool as et


def get_sensor_config():
    config = et.a111.IQServiceConfig()
    config.range_interval = [0.3, 0.6]
    config.update_rate = 80
    config.repetition_mode = et.a111.IQServiceConfig.RepetitionMode.SENSOR_DRIVEN
    return config


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        assert sensor_config.update_rate is not None

        self.f = sensor_config.update_rate
        self.dt = 1 / self.f

        num_hist_points = int(round(self.f * 3))

        self.lp_vel = 0
        self.last_sweep = None
        self.hist_pos = np.zeros(num_hist_points)
        self.sweep_index = 0

    def process(self, data, data_info):
        sweep = data

        n = len(sweep)

        ampl = np.abs(sweep)
        power = np.square(ampl)
        if np.sum(power) > 1e-6:
            com = np.sum(np.arange(n) / n * power) / np.sum(power)  # center of mass
        else:
            com = 0

        if self.sweep_index == 0:
            self.lp_ampl = ampl
            self.lp_com = com
            plot_data = None
        else:
            a = self.alpha(0.1, self.dt)
            self.lp_ampl = a * ampl + (1 - a) * self.lp_ampl
            a = self.alpha(0.25, self.dt)
            self.lp_com = a * com + (1 - a) * self.lp_com

            com_idx = int(self.lp_com * n)
            delta_angle = np.angle(sweep[com_idx] * np.conj(self.last_sweep[com_idx]))
            vel = self.f * 2.5 * delta_angle / (2 * np.pi)

            a = self.alpha(0.1, self.dt)
            self.lp_vel = a * vel + (1 - a) * self.lp_vel

            dp = self.lp_vel / self.f
            self.hist_pos = np.roll(self.hist_pos, -1)
            self.hist_pos[-1] = self.hist_pos[-2] + dp

            hist_len = len(self.hist_pos)
            plot_hist_pos = self.hist_pos - self.hist_pos.mean()
            cut_hist_pos = self.hist_pos[hist_len // 2 :]
            plot_hist_pos_zoom = cut_hist_pos - cut_hist_pos.mean()

            iq_val = np.exp(1j * np.angle(sweep[com_idx])) * self.lp_ampl[com_idx]

            plot_data = {
                "abs": self.lp_ampl,
                "arg": np.angle(sweep),
                "com": self.lp_com,
                "hist_pos": plot_hist_pos,
                "hist_pos_zoom": plot_hist_pos_zoom,
                "iq_val": iq_val,
            }

        self.last_sweep = sweep
        self.sweep_index += 1
        return plot_data

    def alpha(self, tau, dt):
        return 1 - np.exp(-dt / tau)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved
from __future__ import annotations

import typing as t
from enum import Enum

import attrs
import h5py
import numpy as np

from . import core
from .builtin_persistors import ListPersistor
from .registry_persistor import RegistryPersistor


@RegistryPersistor.register_persistor
class ScalarListPersistor(core.Persistor):
    """
    Persists lists of scalar types as a single Dataset
    """

    PRIORITY: t.ClassVar[int] = RegistryPersistor.priority_higher_than(ListPersistor)

    SCALARS = (float, int, bool)

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_list, *type_args = core.unwrap_generic(__type)

        if should_be_list is list:
            (type_arg,) = type_args
            return core.is_subclass(type_arg, cls.SCALARS)
        else:
            return False

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, list):
            raise core.TypeMissmatchError

        (element_type_tree,) = self.type_tree.children.values()

        if any(not isinstance(e, element_type_tree.data) for e in data):
            raise core.TypeMissmatchError

        self.create_own_dataset(data)

    def _load(self) -> t.Any:
        dataset_contents = self.dataset[()]

        self.assert_not_empty(dataset_contents)

        (element_type_tree,) = self.type_tree.children.values()
        element_type = element_type_tree.data

        if not core.is_subclass(element_type, self.SCALARS):
            raise RuntimeError

        return [element_type(e) for e in dataset_contents]


@RegistryPersistor.register_persistor
class RaggedNumpyArrayPersistor(core.Persistor):
    """
    Persists "ragged" (and maybe optional) numpy arrays (list of numpy arrays)
    """

    PRIORITY: t.ClassVar[int] = RegistryPersistor.priority_higher_than(ListPersistor)

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_list, *type_args = core.unwrap_generic(__type)

        if should_be_list is not list:
            return False

        (type_arg,) = type_args

        if core.is_ndarray(type_arg):
            return True

        return core.is_optional(type_arg) and core.is_ndarray(core.optional_arg(type_arg))

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, list):
            raise core.TypeMissmatchError

        if max(np.ndim(array) for array in data) > 1:
            msg = "Only 1-dimensional arrays can be ragged."
            raise core.SaveError(msg)

        try:
            (dtype,) = {arr.dtype for arr in data if arr is not None}
        except ValueError as ve:
            raise core.TypeMissmatchError from ve

        ragged_dtype = h5py.vlen_dtype(dtype)
        num_arrays = len(data)

        ragged_dataset = self.create_own_dataset(
            data=None,
            shape=(num_arrays,),
            dtype=ragged_dtype,
        )

        for i, array in enumerate(data):
            ragged_dataset[i] = array

    def _load(self) -> t.List[t.Any]:
        self.assert_not_empty(self.dataset[()])

        return list(self.dataset[()])


@RegistryPersistor.register_persistor
class TrileanListPersistor(core.Persistor):
    """
    Persists lists of "trileans" (i.e. Optional[bool]) by encoding it as an enum.
    """

    PRIORITY: t.ClassVar[int] = RegistryPersistor.priority_higher_than(ListPersistor)

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_list, *type_args = core.unwrap_generic(__type)

        if should_be_list is list:
            (type_arg,) = type_args
            return type_arg == t.Optional[bool]
        else:
            return False

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, list):
            raise core.TypeMissmatchError

        if any(e not in {True, False, None} for e in data):
            raise core.TypeMissmatchError

        self.create_own_dataset(
            data=[_Trilean.from_object_value(e).stored_value for e in data],
            dtype=_Trilean.dtype(),
        )

    def _load(self) -> t.List[t.Optional[bool]]:
        self.assert_not_empty(self.dataset[()])

        return [_Trilean.from_stored_value(e).object_value for e in self.dataset[()]]


class _Trilean(Enum):
    TRUE = (0, True)
    FALSE = (1, False)
    NONE = (2, None)

    @property
    def stored_value(self) -> int:
        return int(self.value[0])

    @property
    def object_value(self) -> t.Optional[bool]:
        return self.value[1]  # type: ignore[no-any-return]

    @classmethod
    def from_stored_value(cls, value: int) -> t.Any:
        return {member.stored_value: member for member in cls}.get(value)

    @classmethod
    def from_object_value(cls, value: t.Optional[bool]) -> t.Any:
        return {member.object_value: member for member in cls}.get(value)

    @classmethod
    def dtype(cls) -> h5py.Datatype:
        return h5py.enum_dtype(
            {member.name: member.stored_value for member in cls},
            basetype="u1",
        )


@RegistryPersistor.register_persistor
class OptionalFloatListPersistor(core.Persistor):
    """
    Persists a list of Optional[float] by encoding ``None``s as np.inf
    if no np.inf exists in the list
    """

    PRIORITY: t.ClassVar[int] = RegistryPersistor.priority_higher_than(ListPersistor)

    SENTINEL = np.inf

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_list, *type_args = core.unwrap_generic(__type)

        if should_be_list is list:
            (type_arg,) = type_args
            return type_arg == t.Optional[float]
        else:
            return False

    def _save(self, data: t.Any) -> None:
        if any(e == self.SENTINEL for e in data):
            raise core.TypeMissmatchError

        data_with_replaced_nones = [(self.SENTINEL if e is None else e) for e in data]

        self.create_own_dataset(data_with_replaced_nones)

    def _load(self) -> t.List[t.Optional[float]]:
        self.assert_not_empty(self.dataset[()])

        return [(None if e == self.SENTINEL else float(e)) for e in self.dataset[()]]


@RegistryPersistor.register_persistor
class OptionalIntListPersistor(core.Persistor):
    """
    Persists a list of Optional[int] by encoding ``None``s as
    ``np.iinfo(np.int64).min`` (minimum value of int64)
    if the minimum values doesn't exists in the list
    """

    PRIORITY: t.ClassVar[int] = RegistryPersistor.priority_higher_than(ListPersistor)

    SENTINEL = np.iinfo(np.int64).min

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_list, *type_args = core.unwrap_generic(__type)

        if should_be_list is list:
            (type_arg,) = type_args
            return type_arg == t.Optional[int]
        else:
            return False

    def _save(self, data: t.Any) -> None:
        if any(e == self.SENTINEL for e in data):
            raise core.TypeMissmatchError

        data_with_replaced_nones = [(self.SENTINEL if e is None else e) for e in data]

        self.create_own_dataset(data_with_replaced_nones)

    def _load(self) -> t.List[t.Optional[float]]:
        self.assert_not_empty(self.dataset[()])

        return [(None if e == self.SENTINEL else int(e)) for e in self.dataset[()]]


@RegistryPersistor.register_persistor
class EnumListPersistor(core.Persistor):
    """
    Persists a list of (maybe optional) enums by encoding them as integers in a single Dataset
    """

    PRIORITY: t.ClassVar[int] = RegistryPersistor.priority_higher_than(ListPersistor)

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        should_be_list, *type_args = core.unwrap_generic(__type)

        if should_be_list is not list:
            return False
        (type_arg,) = type_args

        if core.is_subclass(type_arg, Enum):
            return True

        return core.is_optional(type_arg) and core.is_subclass(core.optional_arg(type_arg), Enum)

    @property
    def enum_class(self) -> t.Type[Enum]:
        (element_type_tree,) = self.type_tree.children.values()
        element_type = element_type_tree.data

        if core.is_subclass(element_type, Enum):
            return element_type
        elif core.is_optional(element_type) and core.is_subclass(
            core.optional_arg(element_type), Enum
        ):
            return core.optional_arg(element_type)
        else:
            raise RuntimeError

    @property
    def is_optional(self) -> bool:
        (element_type_tree,) = self.type_tree.children.values()
        element_type = element_type_tree.data
        return core.is_optional(element_type)

    def _save(self, data: t.Any) -> None:
        if not isinstance(data, list):
            raise core.TypeMissmatchError

        h5_enum = H5Enum(self.enum_class, include_none=self.is_optional)

        self.create_own_dataset(
            data=[h5_enum.stored_value(e) for e in data],
            dtype=h5_enum.dtype(),
        )

    def _load(self) -> t.List[t.Optional[Enum]]:
        self.assert_not_empty(self.dataset[()])

        h5_enum = H5Enum(self.enum_class, include_none=self.is_optional)

        return [h5_enum.member(e) for e in self.dataset[()]]


@attrs.frozen
class H5Enum:
    enum_class: t.Type[Enum] = attrs.field(validator=attrs.validators.max_len(2**8 - 2))
    include_none: bool = attrs.field(default=False)

    NONE_NAME: t.ClassVar[str] = "NONE"

    @property
    def _name_mapping(self) -> dict[str, int]:
        return {
            **{member.name: index for index, member in enumerate(self.enum_class)},
            **({self.NONE_NAME: len(self.enum_class)} if self.include_none else {}),
        }

    @property
    def _stored_mapping(self) -> dict[int, str]:
        return {index: name for name, index in self._name_mapping.items()}

    def stored_value(self, member: t.Optional[Enum]) -> int:
        return self._name_mapping[self.NONE_NAME if member is None else member.name]

    def member(self, stored_value: int) -> t.Optional[Enum]:
        name = self._stored_mapping[stored_value]
        if name == self.NONE_NAME:
            return None
        else:
            return self.enum_class[self._stored_mapping[stored_value]]

    def dtype(self) -> h5py.Datatype:
        return h5py.enum_dtype(self._name_mapping, basetype="u1")


@RegistryPersistor.register_persistor
class SausageableAttrsPersistor(core.Persistor):
    """
    Sausages attrs classes that are in a list.

    e.g. the list of attrs instances below

    .. codeblock: python
        @attrs.define
        class Foo:
            i: int
            f: float

        [Foo(1, 1.0), Foo(2, 2.0), Foo(3, 3.0), ...]

    is persisted as if ``Foo`` was defined like

    .. codeblock: python
        @attrs.define
        class Foo:
            i: t.List[int]
            f: t.List[float]

        Foo([1, 2, 3, ...], [1.0, 2.0, 3.0, ...])

    which can allow the fields ``i`` and ``f`` to be saved more efficiently.
    """

    PRIORITY: t.ClassVar[int] = RegistryPersistor.priority_higher_than(ListPersistor)

    @classmethod
    def is_applicable(cls, __type: core.TypeLike) -> bool:
        if not core.is_generic(__type):
            return False

        origin, *type_args = core.unwrap_generic(__type)

        if origin is not list:
            return False

        (type_arg,) = type_args
        return attrs.has(type_arg)

    def _save(self, data: t.Any) -> None:
        (attrs_type_tree,) = self.type_tree.children.values()

        if not isinstance(data, list):
            raise core.TypeMissmatchError

        for attribute_name, attribute_type_tree in attrs_type_tree.children.items():
            RegistryPersistor(
                self.require_own_group(),
                attribute_name,
                core.Node(
                    t.List[attribute_type_tree.data],  # type: ignore[name-defined]
                    {attribute_name: attribute_type_tree},
                ),
            ).save([getattr(attrs_instance, attribute_name) for attrs_instance in data])

    def _load(self) -> t.List[attrs.AttrsInstance]:
        (attrs_type_tree,) = self.type_tree.children.values()
        attrs_type = attrs_type_tree.data

        assert attrs.has(attrs_type)

        attribute_lists = {
            attribute_name: RegistryPersistor(
                self.group,
                attribute_name,
                core.Node(
                    t.List[attribute_type_tree.data],  # type: ignore[name-defined]
                    {attribute_name: attribute_type_tree},
                ),
            ).load()
            for attribute_name, attribute_type_tree in attrs_type_tree.children.items()
        }

        lengths = {
            len(attribute_list) for attribute_name, attribute_list in attribute_lists.items()
        }

        (length,) = lengths

        list_of_kwargs = [
            {
                attribute_name.strip("_"): attribute_list[index]
                for attribute_name, attribute_list in attribute_lists.items()
            }
            for index in range(length)
        ]

        return [attrs_type(**kwargs) for kwargs in list_of_kwargs]



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from .buttons import (
    JsonButtonOperations,
    JsonPresentable,
    create_json_save_load_buttons,
    create_json_save_load_buttons_from_type,
)
from .dialogs import PresentationType, PresenterFunc



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import abc
from typing import Optional, Sequence

from PySide6 import QtCore
from PySide6.QtWidgets import QComboBox, QWidget

from acconeer.exptool._core.communication.comm_devices import CommDevice, SerialDevice, USBDevice
from acconeer.exptool.app.new.app_model import AppModel


class CommDeviceComboBox(QComboBox):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        self.app_model = app_model
        app_model.sig_notify.connect(self._on_app_model_update)
        self.currentTextChanged.connect(self._on_change)
        self.setMinimumWidth(175)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        devices = self._get_available_devices()

        view_ports = [self.itemData(i) for i in range(self.count())]

        with QtCore.QSignalBlocker(self):
            if devices != view_ports:
                self.clear()
                for device in devices:
                    self.addItem(device.display_name(), device)

            current_device = self._get_current_device()

            # QComboBox.findData can't be used since since it doesn't
            # always match different but content wise identical objects.
            index = -1
            for i in range(self.count()):
                if self.itemData(i) == current_device:
                    index = i
                    break
            self.setCurrentIndex(index)

            self.setEnabled(self.count() > 0)
            if self.count() == 0:
                self.addItem("No device available")

    @abc.abstractmethod
    def _on_change(self) -> None:
        pass

    @abc.abstractmethod
    def _get_available_devices(self) -> Sequence[CommDevice]:
        pass

    @abc.abstractmethod
    def _get_current_device(self) -> Optional[CommDevice]:
        pass


class SerialPortComboBox(CommDeviceComboBox):
    def _on_change(self) -> None:
        self.app_model.set_serial_connection_device(self.currentData())

    def _get_available_devices(self) -> Sequence[SerialDevice]:
        return self.app_model.available_serial_devices

    def _get_current_device(self) -> Optional[SerialDevice]:
        return self.app_model.serial_connection_device


class USBDeviceComboBox(CommDeviceComboBox):
    def _on_change(self) -> None:
        self.app_model.set_usb_connection_device(self.currentData())

    def _get_available_devices(self) -> Sequence[USBDevice]:
        return self.app_model.available_usb_devices

    def _get_current_device(self) -> Optional[USBDevice]:
        return self.app_model.usb_connection_device



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import itertools
import threading
from typing import Iterator, Optional

import numpy as np

from PySide6 import QtCore
from PySide6.QtGui import QTextDocument
from PySide6.QtWidgets import (
    QDialog,
    QLabel,
    QPushButton,
    QStatusBar,
    QTextBrowser,
    QVBoxLayout,
    QWidget,
)

import acconeer.exptool
from acconeer.exptool.app.new import check_package_outdated, get_latest_changelog
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.utils import get_module_version

from .icons import BUTTON_ICON_COLOR, REFRESH


class FrameCountLabel(QLabel):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Number of received frames")

        app_model.sig_frame_count.connect(self._on_app_model_frame_count)
        self._on_app_model_frame_count(None)

    def _on_app_model_frame_count(self, frame_count: Optional[int]) -> None:
        if frame_count is None:
            css = "color: #888;"
            text = "Frames: -    "
        else:
            css = ""
            text = f"Frames: {frame_count:5}"

        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)


class RateStatsLabel(QLabel):
    _FPS: int = 10

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Reported update rate")

        self.rate = np.nan
        self.rate_warning = False

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_rate_stats.connect(self._on_app_model_rate_stats)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        if np.isnan(self.rate):
            css = "color: #888;"
            text = "     - Hz"
        else:
            css = "color: #FD5200;" if self.rate_warning else ""
            text = f"{self.rate:>6.1f} Hz"

        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)

    def _on_app_model_rate_stats(
        self,
        rate: float,
        rate_warning: bool,
        jitter: float,
        jitter_warning: bool,
    ) -> None:
        self.rate = rate
        self.rate_warning = rate_warning


class JitterStatsLabel(QLabel):
    _FPS: int = 10

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Reported jitter")

        self.jitter = np.nan
        self.jitter_warning = False

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_rate_stats.connect(self._on_app_model_rate_stats)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        if np.isnan(self.jitter):
            css = "color: #888;"
            text = "    - ms"
        else:
            css = "color: #FD5200;" if self.jitter_warning else ""
            text = f"{self.jitter * 1e3:5.1f} ms"

        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)

    def _on_app_model_rate_stats(
        self,
        rate: float,
        rate_warning: bool,
        jitter: float,
        jitter_warning: bool,
    ) -> None:
        self.jitter = jitter
        self.jitter_warning = jitter_warning


class BackendCPUPercentLabel(QLabel):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Client process CPU load")

        app_model.sig_backend_cpu_percent.connect(self._on_app_model_backend_cpu_percent)

        self._on_app_model_backend_cpu_percent(0)

    def _on_app_model_backend_cpu_percent(self, cpu_percent: int) -> None:
        self.setText(f"CPU: {cpu_percent:3}%")

        css = "color: #FD5200;" if cpu_percent >= 85 else ""
        self.setStyleSheet(f"QWidget{{{css}}}")


class RSSVersionLabel(QLabel):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        app_model.sig_notify.connect(self._on_app_model_update)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        if app_model.rss_version is None:
            css = "color: #888;"
            text = "RSS: <not connected>"
            tooltip = ""
        else:
            text = f"RSS: {app_model.rss_version}"

            if app_model.connection_warning:
                css = "background-color: #D78100; color: #e2e2e2;"
                tooltip = app_model.connection_warning
            else:
                css = ""
                tooltip = ""

        self.setToolTip(tooltip)
        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)


class VersionButton(QPushButton):
    sig_payload = QtCore.Signal(
        tuple
    )  # tuple[changelog (str), latest version (str), outdated? (bool)]

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        et_version = get_module_version(acconeer.exptool)
        self._changelog_text: Optional[str] = None

        self.setFlat(True)
        self.setText(f"ET: {et_version}")
        self.setStyleSheet("color: #4d5157;")

        self.clicked.connect(self._on_click)
        self.sig_payload.connect(self._handle_payload)

        self._version_thread = threading.Thread(target=lambda: self._get_payload(et_version))
        self._version_thread.start()

    def _on_click(self) -> None:
        if self._changelog_text is None:
            return

        ChangelogDialog(markdown_text=self._changelog_text).exec()

    def _handle_payload(self, payload: tuple[str, str, bool]) -> None:
        (changelog_text, latest_version, is_outdated) = payload

        if not is_outdated:
            return

        self._changelog_text = changelog_text

        self.setIcon(REFRESH(color="#e2e2e2"))
        self.setStyleSheet(
            f"QPushButton{{ background-color: {BUTTON_ICON_COLOR}; color: #e2e2e2; }}"
        )
        self.setToolTip(
            "There is a new software version available!\n"
            f"The latest version is: {latest_version}. \n"
            "Click to view changelog."
        )

    def _get_payload(self, et_version: str) -> None:
        is_outdated, latest_version = check_package_outdated("acconeer-exptool", et_version)
        self.sig_payload.emit((get_latest_changelog(), latest_version, is_outdated))


class StatusBar(QStatusBar):
    MAX_DISPLAY_TIME_MS = 3000

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self._status_messages: list[str] = []

        app_model.sig_status_file_path.connect(self._update_file_path)
        app_model.sig_status_message.connect(lambda m: self._status_messages.append(m))
        app_model.sig_status_message.connect(self._trigger_display)

        self.message_timer = QtCore.QTimer(self)
        self.message_timer.setInterval(self.MAX_DISPLAY_TIME_MS)
        self.message_timer.setSingleShot(True)
        self.message_timer.timeout.connect(self._trigger_display)

        self.message_widget = QLabel(self)
        self.addWidget(self.message_widget)

        self.file_path_widget = QLabel(self)
        self.addWidget(self.file_path_widget)

        self.addPermanentWidget(FrameCountLabel(app_model, self))
        self.addPermanentWidget(RateStatsLabel(app_model, self))
        self.addPermanentWidget(JitterStatsLabel(app_model, self))
        self.addPermanentWidget(BackendCPUPercentLabel(app_model, self))
        self.addPermanentWidget(RSSVersionLabel(app_model, self))
        self.addPermanentWidget(VersionButton(app_model, self))

        font_families = [
            "Consolas",  # Windows
            "Droid Sans Mono",  # Ubuntu
            "DejaVu Sans Mono",  # Ubuntu, backup
            "SF Mono",  # Mac
        ]
        font_family = ", ".join(f'"{ff}"' for ff in font_families)
        self.setStyleSheet(f"QWidget{{font-family: {font_family};}}")

    def _update_file_path(self, file_path: str, opened: bool) -> None:
        self.file_path_widget.setText(f"Replaying file: <b>{file_path}<\b>" if opened else "")

    def _trigger_display(self) -> None:
        if self.message_timer.isActive():
            return

        if self._status_messages:
            (message, occurances) = next(self._rle_status_messages)

            if occurances > 1:
                self.message_widget.setText(f"{message} (x{occurances})")
            else:
                self.message_widget.setText(message)

            self._status_messages = self._status_messages[occurances:]
            self.message_timer.start(self._message_display_time)
        else:
            self.message_widget.clear()

    @property
    def _rle_status_messages(self) -> Iterator[tuple[str, int]]:
        return (
            (message, len(list(group)))
            for message, group in itertools.groupby(self._status_messages)
        )

    @property
    def _message_display_time(self) -> int:
        num_messages = len(list(self._rle_status_messages))
        return int(self.MAX_DISPLAY_TIME_MS / (1 + 2 ** (num_messages - 5)))


class ChangelogDialog(QDialog):
    def __init__(self, markdown_text: str, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        document = QTextDocument()
        document.setMarkdown(markdown_text)

        text_browser = QTextBrowser(self)
        text_browser.setDocument(document)
        text_browser.setMinimumWidth(round(document.idealWidth() * 1.1))
        text_browser.setMinimumHeight(500)

        layout = QVBoxLayout()

        layout.addWidget(text_browser)

        self.setLayout(layout)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import typing as t

from PySide6.QtWidgets import QDialog, QFileDialog, QHBoxLayout, QWidget

from .preview import PresentationType, PresentationWindow, PresenterFunc


class SaveDialogWithPreview(QDialog):
    def __init__(
        self,
        dialog: QFileDialog,
        model: t.Any,
        presenter: PresenterFunc,
        parent: t.Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)

        self._dialog = dialog
        self._dialog.finished.connect(self.done)

        presentations = {p_type: presenter(model, p_type) for p_type in PresentationType}
        non_none_presentations = {
            p_type: p for p_type, p in presentations.items() if p is not None
        }
        layout = QHBoxLayout()
        layout.addWidget(self._dialog, stretch=1)
        layout.addWidget(
            PresentationWindow(non_none_presentations),
            stretch=1,
        )
        self.setLayout(layout)

    @classmethod
    def get_save_file_name(
        cls,
        model: t.Any,
        presenter: PresenterFunc,
        caption: str = "",
        filter: str = "All (*)",
        options: QFileDialog.Option = QFileDialog.Option.DontUseNativeDialog,
        parent: t.Optional[QWidget] = None,
    ) -> t.Tuple[t.Optional[str], t.Optional[str]]:
        dialog = QFileDialog(caption=caption)
        dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        dialog.setNameFilter(filter)
        dialog.setOptions(options)

        instance = cls(dialog, model, presenter, parent=parent)
        if instance.exec():
            (selected_file,) = instance._dialog.selectedFiles()
            selected_filter = instance._dialog.selectedNameFilter()
            return str(selected_file), selected_filter
        else:
            return None, None



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import abc
from typing import Callable, Generic

import attrs

from acconeer.exptool.a121.algo._base import InputT, MetadataT, ProcessorConfigT, ResultT
from acconeer.exptool.app.new import AppModel, Message, PlotPluginBase, PluginSpecBase

from .backend_plugin import GenericProcessorBackendPluginBase
from .view_plugin import ProcessorViewPluginBase


@attrs.frozen(kw_only=True)
class ProcessorPluginSpec(
    PluginSpecBase, abc.ABC, Generic[InputT, ProcessorConfigT, ResultT, MetadataT]
):
    @abc.abstractmethod
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> GenericProcessorBackendPluginBase[InputT, ProcessorConfigT, ResultT, MetadataT]:
        pass

    @abc.abstractmethod
    def create_view_plugin(self, app_model: AppModel) -> ProcessorViewPluginBase[ProcessorConfigT]:
        pass

    @abc.abstractmethod
    def create_plot_plugin(self, app_model: AppModel) -> PlotPluginBase:
        pass



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
from __future__ import annotations

import logging
import typing as t

import h5py

from . import core
from .registry_persistor import RegistryPersistor


_T = t.TypeVar("_T")
_LOG = logging.getLogger(__name__)


def serialize(
    instance: t.Any, group: h5py.Group, *, override_type: t.Optional[core.TypeLike] = None
) -> None:
    """
    Serialize and save an arbitrary object to the specified group
    """
    type_tree = core.create_type_tree(override_type or type(instance))
    core.sanitize_instance(instance, type_tree)
    RegistryPersistor(group, "./", type_tree).save(instance)


def deserialize(group: h5py.Group, typ: t.Type[_T]) -> _T:
    """
    Load and deserialize an object of type 'typ' from the specified group

    Will raise an exception if anything goes wrong.
    """
    type_tree = core.create_type_tree(typ)
    loaded = RegistryPersistor(group, "./", type_tree).load()
    core.sanitize_instance(loaded, type_tree)

    return loaded  # type: ignore[no-any-return]


def try_deserialize(group: h5py.Group, typ: t.Type[_T]) -> t.Optional[_T]:
    """
    Like deserialize, but returns None instead of an raising an exception when an error occurs.
    """
    try:
        loaded = deserialize(group, typ)
    except Exception as e:
        _LOG.info(
            f"Could not deserialize a {typ.__name__} from {group} due to exception: '{e!s}'."
        )
        return None
    else:
        return loaded


register_persistor = RegistryPersistor.register_persistor


def register_json_presentable(__class: t.Type[core.JsonPresentable]) -> None:
    """
    Register a JSON presentable class (class with 'to_json' & 'from_json')

    Instances of this type will be saved as a JSON string.

    This can be called many times with the same type without repercussions.
    """
    RegistryPersistor.register_persistor(core.Persistor.from_json_presentable(__class))



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import abc
from typing import Dict, List, Optional, Type

from .setup_group import SetupGroup


class PlatformInstall(SetupGroup):
    """Base class and registry for platform install scripts"""

    __registry: Dict[str, Type[PlatformInstall]] = {}

    @classmethod
    @abc.abstractmethod
    def get_key(cls) -> str:
        """Gets the key of the platform.

        This is the same key that gets displayed in the selection menu and is
        used with the `--platform` command line argument.
        """
        pass

    @classmethod
    def from_key(cls, platform_key: str) -> Optional[PlatformInstall]:
        """Constructs a Platform subclass that has been registered"""
        subclass = cls.__registry.get(platform_key)
        if subclass is None:
            return None
        else:
            return subclass()

    @classmethod
    def platforms(cls) -> List[str]:
        """Returns the list of keys that are registered"""
        return list(cls.__registry.keys())

    @classmethod
    def register(cls, subclass: Type[PlatformInstall]) -> Type[PlatformInstall]:
        """Registers a subclass"""
        if not issubclass(subclass, cls):
            msg = f"{subclass.__name__!r} needs to be a subclass of {cls.__name__}."
            raise TypeError(msg)
        cls.__registry[subclass.get_key()] = subclass
        return subclass



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import enum
from typing import Any, Dict, List, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121, opser
from acconeer.exptool.a121._core.entities.configs.config_enums import PRF
from acconeer.exptool.a121._core.utils import (
    is_divisor_of,
    is_multiple_of,
    map_over_extended_structure,
)
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    PERCEIVED_WAVELENGTH,
    AlgoBase,
    AlgoConfigBase,
    PeakSortingMethod,
    calculate_loopback_peak_location,
    get_distance_filter_edge_margin,
)
from acconeer.exptool.a121.algo.touchless_button import (
    MeasurementType,
)
from acconeer.exptool.a121.algo.touchless_button import (
    Processor as TouchlessbuttonProcessor,
)
from acconeer.exptool.a121.algo.touchless_button import (
    ProcessorConfig as TouchlessbuttonProcessorConfig,
)

from ._processors import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
    apply_max_depth_filter,
)


SPARSE_IQ_PPC = 24


@attrs.mutable(kw_only=True)
class DetectorContext(AlgoBase):
    single_sensor_contexts: Dict[int, SingleSensorContext] = attrs.field(factory=dict)
    calibration_session_config: Optional[a121.SessionConfig] = attrs.field(default=None)

    @property
    def sensor_ids(self) -> list[int]:
        return list(self.single_sensor_contexts.keys())


@attrs.mutable(kw_only=True)
class SingleSubsweepContext(AlgoBase):
    mean_sweep: npt.NDArray[np.float64]
    std_sweep: npt.NDArray[np.float64]


@attrs.mutable(kw_only=True)
class SingleSensorContext(AlgoBase):
    subsweep_contexts: List[SingleSubsweepContext] = attrs.field(factory=list)
    reference_temperature: Optional[float] = attrs.field(default=None)
    loopback_peak_location_m: Optional[float] = attrs.field(default=None)


@attributes_doc
@attrs.mutable(kw_only=True)
class DetectorConfig(AlgoConfigBase):
    start_m: float = attrs.field(default=0.15)
    """Start point of measurement interval in meters."""

    end_m: float = attrs.field(default=0.6)
    """End point of measurement interval in meters."""

    step_length: int = attrs.field(default=2)
    """Used to set step length. In unit approx. 2.5 mm."""

    hwaas: int = attrs.field(default=12)
    """Hardware averaging. Higher gives better SNR but increase power
    consumption and lower sweep rate."""

    profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_3, converter=a121.Profile)
    """Profile, 1-5. Higher equals better SNR and lower increase resolution.
    A recommendation is Profile 1 closer than 20 cm and Profile 3 beyond."""

    max_robot_speed: float = attrs.field(default=0.5)
    """Sets the sweep rate after the maximum robot speed in meters per second"""

    sweeps_per_frame: int = attrs.field(default=16)
    """Number of sweeps per frame. The length of the FFT to estimate speed or angle."""

    num_frames_in_recorded_threshold: int = attrs.field(default=50)
    """Number of frames used when calibrating threshold."""

    num_std_threshold: float = attrs.field(default=5)
    """Number of standard deviations added to the threshold."""

    num_mean_threshold: float = attrs.field(default=2)
    """Number of means added to the threshold."""

    update_rate: float = attrs.field(default=50.0)
    """Sets the detector update rate."""

    subsweep_configurations: Optional[List[a121.SubsweepConfig]] = attrs.field(default=None)
    """Optional list of subsweep configurations that over-writes the sensor configuration."""

    peak_sorting_method: PeakSortingMethod = attrs.field(
        default=PeakSortingMethod.CLOSEST,
        converter=PeakSortingMethod,
    )
    """Sorting method of targets."""
    enable_close_proximity_detection: bool = attrs.field(default=False)
    """Enable close proximity detection. Utilizes the first subsweep to control the activity very close to the
     sensor. Extends the result with the flag ´close_proximity_trig´ which is ´True´ if an object
     suddenly appears close to the sensor. If ´subsweep_configurations´ is not set, then the detector
     will automatically add an additional subsweep for the close proximity functionality, otherwise
     the first subsweep will be utilized. """

    dead_reckoning_duration_s: float = attrs.field(default=0.5)
    """Specify the duration (s) of the Kalman filter to perform dead reckoning before stop
    tracking."""

    kalman_sensitivity: float = attrs.field(default=0.5)
    """Specify the sensitivity of the Kalman filter. A higher value yields a more responsive
    filter. A lower value yields a more robust filter."""

    @step_length.validator
    def _validate_step_length(self, _: Any, step_length: int) -> None:
        if step_length is not None and not (
            is_divisor_of(SPARSE_IQ_PPC, step_length) or is_multiple_of(SPARSE_IQ_PPC, step_length)
        ):
            msg = f"step_length must be a divisor or multiple of {SPARSE_IQ_PPC}"
            raise ValueError(msg)

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if self.end_m < self.start_m:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "start_m",
                    "Must be smaller than 'Range end'",
                )
            )

            validation_results.append(
                a121.ValidationError(
                    self,
                    "end_m",
                    "Must be greater than 'Range start'",
                )
            )

        if self.sweeps_per_frame < 2:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sweeps_per_frame",
                    "Must be larger than one for angle estimation.",
                )
            )

        if self.max_robot_speed <= 0.0:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "max_robot_speed",
                    "Must be larger than 0.0.",
                )
            )
        return validation_results


@attrs.frozen(kw_only=True)
class DetectorResult:
    current_velocity: float = attrs.field(default=None)
    processor_results: Dict[int, ProcessorResult] = attrs.field(default=None)
    close_proximity_trig: Optional[Dict[int, bool]] = attrs.field(default=None)


@attrs.mutable(kw_only=True)
class DetectorStatus:
    detector_state: DetailedStatus
    ready_to_start: bool = attrs.field(default=False)


class DetailedStatus(enum.Enum):
    OK = enum.auto()
    END_LESSER_THAN_START = enum.auto()
    SENSOR_IDS_NOT_UNIQUE = enum.auto()
    CONTEXT_MISSING = enum.auto()
    CALIBRATION_MISSING = enum.auto()
    CONFIG_MISMATCH = enum.auto()


class Detector:
    """Obstacle detector
    :param client: Client
    :param sensor_id: Sensor id
    :param detector_config: Detector configuration
    :param context: Detector context
    """

    session_config: a121.SessionConfig
    processor_specs: List[ProcessorConfig]
    context: DetectorContext

    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_ids: list[int],
        detector_config: DetectorConfig,
        context: Optional[DetectorContext] = None,
    ) -> None:
        self.client = client
        self.sensor_ids = sensor_ids
        self.detector_config = detector_config

        self.started = False
        if context is None or not bool(context.single_sensor_contexts):
            self.context = DetectorContext(
                single_sensor_contexts={
                    sensor_id: SingleSensorContext() for sensor_id in self.sensor_ids
                }
            )
            self.detector_status = DetectorStatus(
                detector_state=DetailedStatus.CALIBRATION_MISSING,
                ready_to_start=False,
            )
        else:
            self.context = context
            self.detector_status = DetectorStatus(
                detector_state=DetailedStatus.OK,
                ready_to_start=True,
            )

        # Unless updated, assume that robot moves at max speed
        self.v_current = self.detector_config.max_robot_speed

        self.update_config(self.detector_config)
        self.true_zero_dist_idx = 0

    def _validate_ready_for_calibration(self) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)
        if self.processor_specs is None:
            msg = "Processor specification not defined"
            raise ValueError(msg)
        if self.session_config is None:
            msg = "Session config not defined"
            raise ValueError(msg)

    def calibrate_detector(self) -> None:
        """Run the required detector calibration routines, based on the detector config."""

        self._validate_ready_for_calibration()
        self._calibrate_offset()
        self._calibrate_threshold()

        self.context.calibration_session_config = self.session_config

        self.detector_status.detector_state = DetailedStatus.OK
        self.detector_status.ready_to_start = True

    def start(
        self,
        recorder: Optional[a121.Recorder] = None,
        _algo_group: Optional[h5py.Group] = None,
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        if not self.detector_status.ready_to_start:
            msg = "Not ready to start"
            raise RuntimeError(msg)

        sensor_config = self._get_sensor_config(self.detector_config)

        self.processors: Dict[int, Processor] = {}
        group = []
        for s_id in self.sensor_ids:
            group.append({s_id: sensor_config})

        self.session_config = a121.SessionConfig(
            group,
            extended=True,
            update_rate=self.detector_config.update_rate,
        )

        self.metadata = self.client.setup_session(self.session_config)

        if self.detector_config.enable_close_proximity_detection is True:
            self.close_proximity_processors: Dict[int, TouchlessbuttonProcessor] = {}
            assert isinstance(self.metadata, list)
            (
                self.metadata_first_subsweep,
                self.metadata_rest_subsweep,
            ) = self.split_metadata(self.metadata)
            merged_metadata_first_subsweep = {}
            for m in self.metadata_first_subsweep:
                merged_metadata_first_subsweep.update(m)
            (
                self.sensor_config_first_subsweep,
                self.sensor_config_rest_subsweep,
            ) = self.split_sensor_config(sensor_config)

        for s_id in self.sensor_ids:
            sens_context = self.context.single_sensor_contexts[s_id]

            mean_sweeps: list[npt.NDArray[np.float64]] = []
            std_sweeps: list[npt.NDArray[np.float64]] = []

            if self.detector_config.enable_close_proximity_detection is True:
                obstacle_proc_sensor_config = self.sensor_config_rest_subsweep

                close_proximity_config = TouchlessbuttonProcessorConfig(
                    sensitivity_close=1.9,
                    patience_close=2,
                    calibration_duration_s=1,
                    calibration_interval_s=10,
                    measurement_type=MeasurementType.CLOSE_RANGE,
                )
                self.close_proximity_processors[s_id] = TouchlessbuttonProcessor(
                    sensor_config=self.sensor_config_first_subsweep,
                    metadata=merged_metadata_first_subsweep[s_id],
                    processor_config=close_proximity_config,
                )
            else:
                obstacle_proc_sensor_config = sensor_config

            for subsweep_idx in range(obstacle_proc_sensor_config.num_subsweeps):
                ssc = sens_context.subsweep_contexts[subsweep_idx]
                mean_sweeps.append(ssc.mean_sweep)
                std_sweeps.append(ssc.std_sweep)

            assert sens_context.loopback_peak_location_m is not None
            proc_context = ProcessorContext(
                update_rate=self.detector_config.update_rate,
                mean_sweeps=mean_sweeps,
                std_sweeps=std_sweeps,
                loopback_peak_location_m=sens_context.loopback_peak_location_m,
                reference_temperature=sens_context.reference_temperature,
            )

            pc = ProcessorConfig(
                num_std_treshold=self.detector_config.num_std_threshold,
                num_mean_treshold=self.detector_config.num_mean_threshold,
                peak_sorting_method=self.detector_config.peak_sorting_method,
                max_robot_speed=self.detector_config.max_robot_speed,
                dead_reckoning_duration_s=self.detector_config.dead_reckoning_duration_s,
                sensitivity=self.detector_config.kalman_sensitivity,
            )

            self.processors[s_id] = Processor(
                sensor_config=obstacle_proc_sensor_config,
                processor_config=pc,
                context=proc_context,
            )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("obstacle_detector")

                _record_algo_data(
                    _algo_group,
                    self.sensor_ids,
                    self.detector_config,
                    self.context,
                )

            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.detector_status.ready_to_start = False

        self.started = True

    def get_next(self) -> DetectorResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        results = self.client.get_next()
        assert isinstance(results, list)

        if self.detector_config.enable_close_proximity_detection is True:
            first_subsweep_results, rest_subsweep_results = self.split_result(results)
            close_proximity_results = {}
            assert self.close_proximity_processors is not None
            for r in first_subsweep_results:
                (sens_id,) = r.keys()  # get the only sensor-id in that dict.
                close_result = self.close_proximity_processors[sens_id].process(r[sens_id]).close
                assert close_result is not None
                close_proximity_results[sens_id] = close_result.detection

        else:
            rest_subsweep_results = results
            close_proximity_results = None

        obstacle_proc_results: Dict[int, ProcessorResult] = {}

        for r in rest_subsweep_results:
            sens_id = list(r.keys())[0]  # get the only sensor-id in that dict.
            obstacle_proc_results[sens_id] = self.processors[sens_id].process(r[sens_id])

        return DetectorResult(
            current_velocity=self.v_current,
            processor_results=obstacle_proc_results,
            close_proximity_trig=close_proximity_results,
        )

    def split_metadata(
        self, metadata: List[Dict[int, a121.Metadata]]
    ) -> Tuple[List[Dict[int, a121.Metadata]], List[Dict[int, a121.Metadata]]]:
        first_subsweep_metadata_list = map_over_extended_structure(
            self.first_subsweep_metadata,
            metadata,
        )
        rest_subsweep_metadata_list = map_over_extended_structure(
            self.rest_subsweep_metadata,
            metadata,
        )
        return first_subsweep_metadata_list, rest_subsweep_metadata_list

    def first_subsweep_metadata(self, metadata: a121.Metadata) -> a121.Metadata:
        num_points_first_subsweep = metadata.subsweep_data_length[0]
        frame_data_length = num_points_first_subsweep * metadata.frame_shape[0]
        first_subsweep_metadata = a121.Metadata(
            frame_data_length=frame_data_length,
            sweep_data_length=num_points_first_subsweep,
            subsweep_data_offset=np.array([0]),
            subsweep_data_length=np.array([num_points_first_subsweep]),
            calibration_temperature=metadata.calibration_temperature,
            tick_period=metadata.tick_period,
            base_step_length_m=metadata.base_step_length_m,
            max_sweep_rate=metadata.max_sweep_rate,
            high_speed_mode=metadata.high_speed_mode,
        )
        return first_subsweep_metadata

    def rest_subsweep_metadata(self, metadata: a121.Metadata) -> a121.Metadata:
        num_points_first_subsweep = metadata.subsweep_data_length[0]
        frame_data_length = (
            metadata.frame_data_length - num_points_first_subsweep * metadata.frame_shape[0]
        )
        rest_subsweep_metadata = a121.Metadata(
            frame_data_length=frame_data_length,
            sweep_data_length=metadata.frame_shape[0] - num_points_first_subsweep,
            subsweep_data_offset=metadata.subsweep_data_offset[1:] - num_points_first_subsweep,
            subsweep_data_length=metadata.subsweep_data_length[1:],
            calibration_temperature=metadata.calibration_temperature,
            tick_period=metadata.tick_period,
            base_step_length_m=metadata.base_step_length_m,
            max_sweep_rate=metadata.max_sweep_rate,
            high_speed_mode=metadata.high_speed_mode,
        )
        return rest_subsweep_metadata

    def split_sensor_config(
        self, sensor_config: a121.SensorConfig
    ) -> Tuple[a121.SensorConfig, a121.SensorConfig]:
        first_subsweep_sensor_config = a121.SensorConfig(
            subsweeps=[sensor_config.subsweeps[0]],
            sweeps_per_frame=sensor_config.sweeps_per_frame,
            sweep_rate=sensor_config.sweep_rate,
            inter_sweep_idle_state=sensor_config.inter_sweep_idle_state,
            inter_frame_idle_state=sensor_config.inter_frame_idle_state,
        )
        rest_subsweep_sensor_config = a121.SensorConfig(
            subsweeps=sensor_config.subsweeps[1:],
            sweeps_per_frame=sensor_config.sweeps_per_frame,
            sweep_rate=sensor_config.sweep_rate,
            inter_sweep_idle_state=sensor_config.inter_sweep_idle_state,
            inter_frame_idle_state=sensor_config.inter_frame_idle_state,
        )

        return first_subsweep_sensor_config, rest_subsweep_sensor_config

    def split_result(
        self, results: List[Dict[int, a121.Result]]
    ) -> Tuple[List[Dict[int, a121.Result]], List[Dict[int, a121.Result]]]:
        first_subsweep_results = map_over_extended_structure(
            self.first_subsweep_result,
            results,
        )
        rest_subsweep_results = map_over_extended_structure(self.rest_subsweep_result, results)

        return first_subsweep_results, rest_subsweep_results

    def first_subsweep_result(self, result: a121.Result) -> a121.Result:
        num_points_first_subsweep = result._context.metadata.subsweep_data_length[0]
        first_subsweep_result = a121.Result(
            data_saturated=result.data_saturated,
            frame_delayed=result.frame_delayed,
            calibration_needed=result.calibration_needed,
            temperature=result.temperature,
            frame=result._frame[:, 0:num_points_first_subsweep],
            tick=result.tick,
            context=a121._core.entities.ResultContext(
                metadata=self.first_subsweep_metadata(result._context.metadata),
                ticks_per_second=result._context.ticks_per_second,
            ),
        )
        return first_subsweep_result

    def rest_subsweep_result(self, result: a121.Result) -> a121.Result:
        num_points_first_subsweep = result._context.metadata.subsweep_data_length[0]
        rest_subsweep_result = a121.Result(
            data_saturated=result.data_saturated,
            frame_delayed=result.frame_delayed,
            calibration_needed=result.calibration_needed,
            temperature=result.temperature,
            frame=result._frame[:, num_points_first_subsweep:],
            tick=result.tick,
            context=a121._core.entities.ResultContext(
                metadata=self.rest_subsweep_metadata(result._context.metadata),
                ticks_per_second=result._context.ticks_per_second,
            ),
        )
        return rest_subsweep_result

    def update_robot_speed(self, v_current: float) -> None:
        self.v_current = v_current

    @staticmethod
    def get_obstacle_angle(v_current: float, v_measured: float) -> float:
        """
        Convert the measured speed of an object to an angle.

        The convention is that a robot moving forward has a positive robot speed
        and an object moving towards from the sensor has a negative measured speed.
        Therefore, for an object straight in front of the sensor, v_measured is
        approximately -v_current and the obstacle angle is positive and close
        to zero.
        """
        return float(np.arccos(-v_measured / v_current))

    def stop(self) -> Any:
        """Stops the measurement session."""
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()

        self.detector_status.ready_to_start = True

        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result

    def update_config(self, config: DetectorConfig) -> None:
        """Updates the session config and processor specification based on the detector
        configuration."""
        (
            self.session_config,
            self.processor_specs,
        ) = self._detector_to_session_config_and_processor_specs(
            config=config, sensor_ids=self.sensor_ids
        )

        if self.context.calibration_session_config != self.session_config:
            self.detector_status.detector_state = DetailedStatus.CALIBRATION_MISSING
            self.detector_status.ready_to_start = False

    @classmethod
    def _get_sensor_config(cls, detector_config: DetectorConfig) -> a121.SensorConfig:
        if detector_config.subsweep_configurations is None:
            subsweeps = []
            if detector_config.enable_close_proximity_detection is True:
                close_proximity_subsweep = a121.SubsweepConfig(
                    start_point=8,
                    num_points=2,
                    step_length=6,
                    profile=a121.Profile.PROFILE_1,
                    hwaas=40,
                    prf=PRF.PRF_15_6_MHz,
                )
                subsweeps.append(close_proximity_subsweep)

            start_p = int(detector_config.start_m / APPROX_BASE_STEP_LENGTH_M)
            point_margin = get_distance_filter_edge_margin(
                detector_config.profile, detector_config.step_length
            )
            start_p = start_p - point_margin * detector_config.step_length

            num_p = int(
                (detector_config.end_m - detector_config.start_m)
                / (APPROX_BASE_STEP_LENGTH_M * detector_config.step_length)
            )
            num_p = num_p + 2 * point_margin

            subsweeps.append(
                a121.SubsweepConfig(
                    start_point=start_p,
                    num_points=num_p,
                    step_length=detector_config.step_length,
                    profile=detector_config.profile,
                    hwaas=detector_config.hwaas,
                    prf=PRF.PRF_15_6_MHz,
                )
            )

        else:
            subsweeps = detector_config.subsweep_configurations

        for subsweep in subsweeps:
            subsweep.phase_enhancement = True

        sweep_rate: Optional[float] = detector_config.max_robot_speed / (PERCEIVED_WAVELENGTH / 2)
        if detector_config.max_robot_speed / (PERCEIVED_WAVELENGTH / 2) <= 0.0:
            sweep_rate = None

        return a121.SensorConfig(
            sweeps_per_frame=detector_config.sweeps_per_frame,
            sweep_rate=sweep_rate,
            inter_sweep_idle_state=a121.IdleState.READY,
            inter_frame_idle_state=a121.IdleState.READY,
            subsweeps=subsweeps,
        )

    @classmethod
    def _detector_to_session_config_and_processor_specs(
        cls, config: DetectorConfig, sensor_ids: list[int]
    ) -> Tuple[a121.SessionConfig, list[ProcessorConfig]]:
        processor_spec = ProcessorConfig(
            num_std_treshold=config.num_std_threshold,
            num_mean_treshold=config.num_mean_threshold,
        )

        sensor_config = cls._get_sensor_config(config)
        return (
            a121.SessionConfig({s_id: sensor_config for s_id in sensor_ids}, extended=True),
            [processor_spec] * len(sensor_ids),
        )

    def _calibrate_offset(self) -> None:
        """Estimates sensor offset error based on loopback measurement."""

        self._validate_ready_for_calibration()

        sensor_config = a121.SensorConfig(
            start_point=-30,
            num_points=50,
            step_length=1,
            profile=a121.Profile.PROFILE_1,
            hwaas=64,
            sweeps_per_frame=1,
            enable_loopback=True,
            phase_enhancement=True,
        )

        # Sensors running simultaneously, profile 1 and direct leakage give low risk of cross talk.
        session_config = a121.SessionConfig(
            {sensor_id: sensor_config for sensor_id in self.sensor_ids},
            extended=True,
        )

        self.client.setup_session(session_config)
        self.client.start_session()
        extended_result = self.client.get_next()
        self.client.stop_session()

        assert isinstance(extended_result, list)

        for sensor_id, context in self.context.single_sensor_contexts.items():
            context.loopback_peak_location_m = calculate_loopback_peak_location(
                extended_result[0][sensor_id], sensor_config
            )

    def _calibrate_threshold(self) -> None:
        self._validate_ready_for_calibration()

        sensor_config = self._get_sensor_config(self.detector_config)

        if self.detector_config.enable_close_proximity_detection:
            _, sensor_config = self.split_sensor_config(
                sensor_config
            )  # Don't use first subsweep for calibration

        session_config = a121.SessionConfig(
            {sensor_id: sensor_config for sensor_id in self.sensor_ids}, extended=True
        )
        self.client.setup_session(session_config)

        result_list: list[dict[int, a121.Result]] = []
        self.client.start_session()
        for _ in range(self.detector_config.num_frames_in_recorded_threshold):
            extended_result = self.client.get_next()
            assert isinstance(extended_result, list)
            assert all(isinstance(r, dict) for r in extended_result)
            result_list.append({i: r for d in extended_result for i, r in d.items()})
        self.client.stop_session()

        # Processor used for applying the depth filter
        processor = Processor(
            sensor_config=sensor_config,
            processor_config=ProcessorConfig(),
            context=ProcessorContext(update_rate=self.detector_config.update_rate),
        )

        for s_id in self.sensor_ids:
            results: list[a121.Result] = [r[s_id] for r in result_list]
            depth_filtered_data = [processor.apply_depth_filter(result) for result in results]
            threshold_temp = np.mean([r.temperature for r in results])

            self.context.single_sensor_contexts[s_id].subsweep_contexts = []
            for subsweep_idx in range(sensor_config.num_subsweeps):
                subsweep_filtered_data = np.array(
                    [dfd[subsweep_idx] for dfd in depth_filtered_data]
                )

                abs_mean = np.abs(
                    np.mean(subsweep_filtered_data, axis=(0, 1))
                )  # Over the frame- and sweep dimension
                std = np.std(subsweep_filtered_data, axis=(0, 1))

                subsweep_mean = apply_max_depth_filter(
                    abs_mean, sensor_config.subsweeps[subsweep_idx]
                )

                ssc = SingleSubsweepContext(mean_sweep=subsweep_mean, std_sweep=std)
                self.context.single_sensor_contexts[s_id].subsweep_contexts.append(ssc)

            self.context.single_sensor_contexts[s_id].reference_temperature = float(threshold_temp)

    @classmethod
    def get_detector_status(
        cls,
        config: DetectorConfig,
        context: DetectorContext,
        sensor_ids: list[int],
    ) -> DetectorStatus:
        """Returns the detector status along with the detector state."""

        if config.end_m < config.start_m:
            return DetectorStatus(
                detector_state=DetailedStatus.END_LESSER_THAN_START,
                ready_to_start=False,
            )

        if len(sensor_ids) != len(set(sensor_ids)):
            return DetectorStatus(
                detector_state=DetailedStatus.SENSOR_IDS_NOT_UNIQUE,
                ready_to_start=False,
            )

        if context.single_sensor_contexts is None:
            return DetectorStatus(
                detector_state=DetailedStatus.CONTEXT_MISSING,
                ready_to_start=False,
            )

        (
            session_config,
            _,
        ) = cls._detector_to_session_config_and_processor_specs(
            config=config, sensor_ids=sensor_ids
        )
        # Offset calibration is always performed as a part of the detector calibration process.
        # Use this as indication whether detector calibration has been performed.
        calibration_missing = np.any(
            [
                ctx.loopback_peak_location_m is None
                for ctx in context.single_sensor_contexts.values()
            ]
        )
        config_mismatch = (
            context.calibration_session_config != session_config
            or context.sensor_ids != sensor_ids
        )

        if calibration_missing:
            detector_state = DetailedStatus.CALIBRATION_MISSING
        elif config_mismatch:
            detector_state = DetailedStatus.CONFIG_MISMATCH
        else:
            detector_state = DetailedStatus.OK

        return DetectorStatus(
            detector_state=detector_state,
            ready_to_start=(detector_state == DetailedStatus.OK),
        )


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_ids: list[int],
    config: DetectorConfig,
    context: DetectorContext,
) -> None:
    algo_group.create_dataset(
        "sensor_ids",
        data=sensor_ids,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "detector_config", config.to_json())

    context_group = algo_group.create_group("context")
    opser.serialize(context, context_group)


def _load_algo_data(
    algo_group: h5py.Group,
) -> Tuple[list[int], DetectorConfig, DetectorContext]:
    sensor_id = algo_group["sensor_ids"][()].tolist()
    config = DetectorConfig.from_json(algo_group["detector_config"][()])

    context_group = algo_group["context"]
    context = opser.deserialize(context_group, DetectorContext)

    return sensor_id, config, context



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
from typing import Any

import attrs
import h5py

from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset


@attrs.frozen(kw_only=True)
class SensorCalibration:
    """SensorCalibration

    Represents the RSS ``cal_result`` and ``cal_info``.
    """

    temperature: int = attrs.field()
    """The calibration temperature"""

    data: str = attrs.field()
    """The calibration data"""

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> SensorCalibration:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> SensorCalibration:
        return cls.from_dict(json.loads(json_str))

    def to_h5(self, group: h5py.Group) -> None:
        group.create_dataset("temperature", data=self.temperature, track_times=False)
        _create_h5_string_dataset(group, "data", self.data)

    @classmethod
    def from_h5(cls, group: h5py.Group) -> SensorCalibration:
        return cls.from_dict(
            {
                "temperature": group["temperature"][()],
                "data": group["data"][()].decode("utf-8"),
            }
        )



# Copyright (c) Acconeer AB, 2022-2026
# All rights reserved

from __future__ import annotations

import argparse
import getpass
import re
import tempfile
import time
import typing as t

from acconeer.exptool._core.communication import comm_devices
from acconeer.exptool._core.communication.links.usb_link import UsbPortError
from acconeer.exptool.flash._bin_fetcher import (
    BIN_FETCH_PROMPT,
    clear_cookies,
    download,
    get_content,
    get_cookies,
    is_redirected_to_login_page,
    login,
    save_cookies,
)
from acconeer.exptool.flash._bootloader_tool import BootloaderTool
from acconeer.exptool.flash._flash_exception import FlashException
from acconeer.exptool.flash._products import (
    EVK_TO_PRODUCT_MAP,
    EVKS_THAT_HAVE_FW_ON_DEVSITE,
    PRODUCT_NAME_TO_FLASH_MAP,
    PRODUCT_PID_TO_FLASH_MAP,
)

from ._dev_license import DevLicense
from ._dev_license_tui import DevLicenseTuiDialog


def _query_yes_no(question: str, default: str = "yes") -> bool:
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise FlashException("invalid default answer: '%s'" % default)

    while True:
        print(question + prompt, end="", flush=True)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def _flasher_progress_callback(progress: int, end: bool = False) -> None:
    print(f"\rProgress: [{progress}/100]", end="", flush=True)
    if end:
        print()


def flash_image(
    image_path: str,
    flash_device: comm_devices.CommDevice,
    device_name: str,
    progress_callback: t.Callable[[int, bool], t.Any] = lambda *_: None,
) -> None:
    if flash_device:
        serial_device_name = device_name or flash_device.name
        if serial_device_name is None:
            msg = "Unknown device type"
            raise FlashException(msg)
        serial_device_name = serial_device_name.upper()

        if (
            isinstance(flash_device, comm_devices.USBDevice)
            and flash_device.pid in PRODUCT_PID_TO_FLASH_MAP
        ):
            bootloader = PRODUCT_PID_TO_FLASH_MAP[flash_device.pid]()
            bootloader.flash(
                flash_device, serial_device_name, image_path, progress_callback=progress_callback
            )
        elif (
            isinstance(flash_device, comm_devices.SerialDevice)
            and serial_device_name in PRODUCT_NAME_TO_FLASH_MAP
        ):
            bootloader = PRODUCT_NAME_TO_FLASH_MAP[serial_device_name]()
            bootloader.flash(flash_device, serial_device_name, image_path, progress_callback)
        else:
            msg = f"No flash support device {str(flash_device)}"
            raise FlashException(msg)
    else:
        msg = "No devices found, try connecting a device before flashing."
        raise FlashException(msg)


def _is_usb_device(name: str) -> bool:
    return issubclass(PRODUCT_NAME_TO_FLASH_MAP[name], BootloaderTool)


def _get_flash_device_from_args(
    args: argparse.Namespace,
) -> t.Union[comm_devices.SerialDevice, comm_devices.USBDevice]:
    device_name: str = args.device
    port = args.port if hasattr(args, "port") else None
    serial_number = args.serial_number if hasattr(args, "serial_number") else None
    flash_device: t.Optional[t.Union[comm_devices.SerialDevice, comm_devices.USBDevice]] = None

    if device_name == "XM125":
        device_name = "XE125"
    elif device_name == "XM126":
        device_name = "XB122"

    # Handle serial port devices
    if not _is_usb_device(device_name):
        flash_device = None
        for serial_device in comm_devices.get_serial_devices():
            if isinstance(serial_device, comm_devices.SerialDevice) and port == serial_device.port:
                flash_device = serial_device
                break
        if flash_device is None:
            msg = f"Port '{port}' could not be found\n"
            raise FlashException(msg)
        return flash_device

    # Handle USB devices
    found_usb_devices = []
    for usb_device in comm_devices.get_usb_devices():
        if device_name is not None and usb_device.name != device_name:
            # Device name did not match
            continue
        elif serial_number is not None and usb_device.serial != serial_number:
            # Serial number did not match
            continue
        else:
            found_usb_devices.append(usb_device)

    if len(found_usb_devices) == 0:
        # Handle Serial Port USB devices (Bootloader)
        flash_device = None
        for serial_device in comm_devices.get_serial_devices():
            if serial_device.name == device_name:
                return serial_device

        if serial_number is not None:
            msg = f"No {device_name} device with serial={serial_number} could be found\n"
        else:
            msg = f"No {device_name} device could be found\n"
        raise FlashException(msg)
    elif len(found_usb_devices) > 1:
        print(f"Found multiple {device_name} devices:")
        print("".join([f" - {dev}\n" for dev in found_usb_devices]))
        msg = "Use --serial-number to select device to be flashed"
        raise FlashException(msg)

    flash_device = found_usb_devices[0]
    print(f"Flashing {flash_device}")

    return flash_device


def get_flash_known_devices() -> list[str]:
    known_devices = []
    for _, item in EVK_TO_PRODUCT_MAP.items():
        if item not in known_devices:
            known_devices.append(item)
    return known_devices


def get_flash_download_name(device: comm_devices.CommDevice, device_name: t.Optional[str]) -> str:
    name = device_name or device.name
    if name in EVK_TO_PRODUCT_MAP:
        return EVK_TO_PRODUCT_MAP[name]
    msg = f"Unknown device {name}"
    raise FlashException(msg)


def get_boot_description(
    flash_device: comm_devices.CommDevice,
    device_name: str,
) -> t.Optional[str]:
    flash_device_name = device_name or flash_device.name
    product = None

    if flash_device_name in EVK_TO_PRODUCT_MAP:
        product = EVK_TO_PRODUCT_MAP[flash_device_name]

    if product in PRODUCT_NAME_TO_FLASH_MAP:
        return PRODUCT_NAME_TO_FLASH_MAP[product].get_boot_description(product)

    return None


def get_post_dfu_description(
    flash_device: comm_devices.CommDevice,
    device_name: str,
) -> t.Optional[str]:
    flash_device_name = device_name or flash_device.name
    product = None

    if flash_device_name in EVK_TO_PRODUCT_MAP:
        product = EVK_TO_PRODUCT_MAP[flash_device_name]

    if product in PRODUCT_NAME_TO_FLASH_MAP:
        return PRODUCT_NAME_TO_FLASH_MAP[product].get_post_dfu_description(product)

    return None


def is_fw_downloadable(device_name: str) -> bool:
    device_name_upper = device_name.upper()
    if device_name_upper in EVK_TO_PRODUCT_MAP:
        return EVK_TO_PRODUCT_MAP[device_name_upper] in EVKS_THAT_HAVE_FW_ON_DEVSITE
    return False


def _fetch_and_flash(args: argparse.Namespace) -> None:
    flash_device = _get_flash_device_from_args(args)

    cookies = get_cookies()
    new_login = False
    login_succeed = False
    while login_succeed is False:
        if cookies is None:
            print("\n" + BIN_FETCH_PROMPT)
            email = input("\nEmail: ")
            password = getpass.getpass("Password: ")
            new_login = True
            cookies = login(email, password)

        print("\nLogging in... ", end="", flush=True)
        login_succeed, session, page = get_content(cookies)
        if not login_succeed:
            cookies = None
            print("[Error]")

            if is_redirected_to_login_page(page):
                print(
                    "Request failed. Try downloading the image directly "
                    "from https://developer.acconeer.com"
                )

        else:
            print("[OK]")

    if new_login and _query_yes_no(
        "\nWe use cookies to optimize the service of our applications."
        "\n\nCookie Policy: https://acconeer.com/cookie-policy-eu/"
        "\nPrivacy Statement: https://developer.acconeer.com/privacy-policy/"
        "\n\nSelecting yes to the following question will store a cookie on "
        "your computer:"
        "\nRemember me?"
    ):
        save_cookies(cookies)

    try:
        with tempfile.TemporaryDirectory() as tmp_dir:
            print("Preparing license agreement... ", end="", flush=True)
            license = DevLicense()
            license.load()
            print("[OK]")

            time.sleep(1)
            DevLicenseTuiDialog.run(log="textual.log", license=license)
            license_accepted = DevLicenseTuiDialog.get_accept()

            if license_accepted:
                device_name = get_flash_download_name(flash_device, args.device)
                print(f"Downloading {device_name} image file... ", end="", flush=True)
                bin_path, version = download(
                    session=session,
                    cookies=cookies,
                    path=tmp_dir,
                    device=device_name,
                )
                print(f"[OK] (version: {version})")

                boot_description = get_boot_description(flash_device, device_name)
                if boot_description:
                    boot_description = re.sub("<li>", " - ", boot_description)
                    boot_description = re.sub("</[p|li]+?>", "\n", boot_description)
                    boot_description = re.sub("<[^<]+?>", "", boot_description)
                    print("\n\n" + boot_description)
                if _query_yes_no("Proceed to flashing you device?"):
                    print("Flashing...")
                    flash_image(
                        bin_path,
                        flash_device,
                        device_name=device_name,
                        progress_callback=_flasher_progress_callback,
                    )
                post_dfu_description = get_post_dfu_description(flash_device, device_name)
                if post_dfu_description:
                    post_dfu_description = re.sub("<li>", " - ", post_dfu_description)
                    post_dfu_description = re.sub("</[p|li]+?>", "\n", post_dfu_description)
                    post_dfu_description = re.sub("<[^<]+?>", "", post_dfu_description)
                    print("\n\n" + post_dfu_description)
            else:
                print("You need to accept the license agreement to download the image file.")
    finally:
        session.close()


def _flash(args: argparse.Namespace) -> None:
    flash_device = _get_flash_device_from_args(args)
    flash_image(
        args.image,
        flash_device,
        device_name=args.device,
        progress_callback=_flasher_progress_callback,
    )


def _add_common_options(parser: argparse.ArgumentParser) -> None:
    image_group = parser.add_mutually_exclusive_group(required=True)
    image_group.add_argument("--image", "-i", dest="image", help="Image file to flash")
    image_group.add_argument(
        "--fetch",
        "-f",
        dest="fetch",
        action="store_true",
        help=(
            "Fetch latest image file from the web and flash it. "
            "Requires an Acconeer Developer Account."
        ),
    )


def _add_usb_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--serial-number",
        "-s",
        dest="serial_number",
        help="Device serial number. Only needed if multiple devices are connected.",
        type=str,
    )
    _add_common_options(parser)


def _add_serial_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--port",
        "-p",
        dest="port",
        help="Serial port",
        type=str,
        required=True,
    )
    _add_common_options(parser)


def main() -> None:
    parser = argparse.ArgumentParser(description="Image Flasher")
    parser.prog = "python -m acconeer.exptool.flash"
    action_subparsers = parser.add_subparsers(
        help="sub-command help", dest="operation", required=True
    )
    action_subparsers.add_parser("list", help="List connected devices")
    action_subparsers.add_parser("clear", help="Clears saved login session")
    flash_subparser = action_subparsers.add_parser("flash", help="Flash device")
    device_subparsers = flash_subparser.add_subparsers(
        help="sub-command help", dest="device", required=True
    )

    for key in PRODUCT_NAME_TO_FLASH_MAP:
        if _is_usb_device(key):
            usb_subparser = device_subparsers.add_parser(key, help=f"Flash {key} device")
            _add_usb_options(usb_subparser)
        else:
            serial_subparser = device_subparsers.add_parser(key, help=f"Flash {key} device")
            _add_serial_options(serial_subparser)

    args = parser.parse_args()

    if args.operation == "list":
        usb_devices = comm_devices.get_usb_devices()
        serial_devices = comm_devices.get_serial_devices()
        if not usb_devices and not serial_devices:
            print("No devices available")
        else:
            if usb_devices:
                print("== USB Devices ==")
                print("".join([f" - {dev}\n" for dev in usb_devices]))
            if serial_devices:
                print("== Serial Devices ==")
                print("".join([f" - {dev}\n" for dev in serial_devices]))
        return
    elif args.operation == "clear":
        clear_cookies()
        print("Saved login session cookie deleted!\n")
    elif args.operation == "flash":
        try:
            if args.fetch:
                if is_fw_downloadable(args.device):
                    _fetch_and_flash(args)
                else:
                    print(f"The image file for {args.device} is not available for download")
            elif args.image:
                _flash(args)
        except FlashException as e:
            print(f"\n{str(e)}\n")
        except UsbPortError as e:
            print(f"\n{str(e)}\n")



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from acconeer.exptool._core.int_16_complex import (
    INT_16_COMPLEX,
    complex_array_to_int16_complex,
    int16_complex_array_to_complex,
)

from .configs import (
    PRF,
    IdleState,
    Profile,
    SensorConfig,
    SessionConfig,
    SubsweepConfig,
)
from .containers import (
    Metadata,
    PersistentRecord,
    Record,
    RecordException,
    Result,
    ResultContext,
    SensorCalibration,
    SensorInfo,
    ServerInfo,
    SessionRecord,
    StackedResults,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import typing as t

import numpy as np
import numpy.typing as npt

from .metadata import Metadata


T = t.TypeVar("T", bound=np.generic)


def get_subsweeps_from_frame(frame: npt.NDArray[T], metadata: Metadata) -> list[npt.NDArray[T]]:
    """Gets the subsweeps from a frame (2D, (<sweeps>, <data points>))
    or frames (3D, (<frames>, <sweeps>, <data points>)).
    This is done by slicing the innermost dimension (the <data points> dimension) for both cases.
    """
    offsets = metadata.subsweep_data_offset
    lengths = metadata.subsweep_data_length
    return [frame[..., o : o + l] for o, l in zip(offsets, lengths)]
